import os

import paramiko, json, logging, pyodbc, oracledb, psycopg2, mariadb, mysql.connector, pytz, subprocess, sys
from socket import getfqdn
from winrm import Session
from io import StringIO
from types import SimpleNamespace
from django.utils import timezone
import re, time
import dns.resolver
from clusterCore.exceptions import *
from pymongo import MongoClient
from clusterCore.api.serializer import c123_HostSerializer, object_to_json, PRIMARY_CREATING, PRIMARY_CREATION_FAILED, c123_inventory_serializer
from clusterCore.models import Task, Profile, CustomerKey, c123_node, c123_database, c123_host, TaskLog, ClusterSizing, SystemConfig, c123_inventory, OracleClusterSizing
from django.contrib.auth.models import User
from elasticsearch import Elasticsearch
from redis import Redis
from django.db import connection
from django.urls import reverse
from django.http import JsonResponse
from django.db.models import Sum
from django.core.mail import EmailMessage
from django.template import Template, Context
from django_apscheduler.util import close_old_connections
from dateutil.relativedelta import relativedelta
from datetime import datetime
from apscheduler.triggers.cron import CronTrigger
from uuid import UUID
from clusterCore.enums import DBType, ClusterRoles, OperatingSystem, Severity, Status, Inventory, Schedule, TaskState, Timeout, Node_Size_Allowed
TASK_CERTIFICATE_FILES_DIR = '/123cluster/certs'
SHELL_SAFE_TRANSFORM = str.maketrans({":": "_", ".": "_"})

unlicense_notice = '''
Dear {{user}}

We would like to inform you that the cluster {{cluster}} is currently having nodes without full license. To ensure uninterrupted service and access to all functionalities, we kindly request you to acquire and apply the appropriate license {{within_time}}.
Nodes that require attention:
{{nodes}}
If you have any questions or need further assistance regarding licensing, please don't hesitate to contact our support team.

Thank you for your attention to this matter.

Best regards,
{{company}}
'''


def test_proxmox_connection(hostname, proxmox_username, proxmox_token_name, proxmox_token_value, verify_ssl=False):
    """
        Test if can connect to proxmox host.

        Args:
            hostname (str): Name of host to connect.
            proxmox_username(str): Username to connect as.
            proxmox_token_name(str): Name of proxmox authentication token.
            proxmox_token_value(str): Value of proxmox authentication token.
            verify_ssl(str): Whether to verify ssl connection or not

        Returns:
            bool: Status of connection.
    """
    try:
       from proxmoxer import ProxmoxAPI
       proxmox = ProxmoxAPI(hostname, user=proxmox_username, token_name=proxmox_token_name, token_value=proxmox_token_value, verify_ssl=verify_ssl)
       proxmox.cluster.status.get()
       return True
    except:
        return False


def get_random_available_ip(ip_range, used_ips):
    """
    Pick a random available IP from a given range that is not in the used IP list.

    Args:
        ip_range (str): The range of IPs in CIDR notation (e.g., '172.16.0.0/24').
        used_ips (list): A list of IPs that are already in use.

    Returns:
        str: A randomly selected available IP, or None if no IPs are available.
    """
    import ipaddress
    import random

    # Parse the IP range
    network = ipaddress.ip_network(ip_range, strict=False)

    # Filter available IPs
    available_ips = [str(ip) for ip in network.hosts() if str(ip) not in used_ips]

    # Return a random IP from the available ones, or None if none are available
    return random.choice(available_ips) if available_ips else None

def get_fully_qualified_domain_name(hostname):
    """return fully qualified domain name to use for connections to host"""
    try:
        return (False, getfqdn(hostname))
    except Exception as e:
        return (True, str(e))


def get_backup_description(hours, minutes, seconds, days_of_week = None, days_of_month = None):
    description = ""
    if hours >= 12:
        meridian = "PM"
        if hours > 12:
            hours -= 12
    else:
        meridian = "AM"
        if hours == 0:
            hours = 12
    if not days_of_week and not days_of_month:
        description = "daily"
    if days_of_month:
        days_of_month = [f"{d}{'th' if 11 <= d % 100 <= 13 else {1: 'st', 2: 'nd', 3: 'rd'}.get(d % 10, 'th')}" for d in days_of_month]

        description = f"every {', '.join(days_of_month)} day"
    if days_of_week:
      full_weekdays = {
          'mon': 'Mon',
          'tue': 'Tue',
          'wed': 'Wed',
          'thu': 'Thu',
          'fri': 'Fri',
          'sat': 'Sat',
          'sun': 'Sun'
      }

      days_of_week = [full_weekdays[d] for d in days_of_week]
      description = f"{description} on {'all weekdays' if len(days_of_week) == 7 else ', '.join(days_of_week)}"

    description = f"{description} {'at {:02d}:{:02d}:{:02d} {}'.format(hours, minutes, seconds, meridian)}"
    return description




def  get_avail_change_date(node):
        from django.conf import settings
        app_timezone = pytz.timezone(settings.TIME_ZONE)
        now = timezone.now().astimezone(app_timezone)
        avail_change_raw = node.ex_params.get('avail_change_raw')
        if not avail_change_raw:
            avail_change_raw = now
            node.ex_params['avail_change_raw'] = avail_change_raw.strftime("%Y-%m-%d %H:%M:%S")
            node.save()
            return "1 days"

        avail_change_raw = datetime.strptime(avail_change_raw,"%Y-%m-%d %H:%M:%S").astimezone(app_timezone)
        passed_delta = relativedelta(now, avail_change_raw)
        node.ex_params['avail_change_raw'] = now.strftime("%Y-%m-%d %H:%M:%S")
        node.save()
        if passed_delta.months < 1:
            return f'{passed_delta.days if passed_delta.days > 0 else 0} days'
        return avail_change_raw.strftime("%Y-%m-%d")

def spread_licenses_by_type(db_type, summary_count):
    targets = c123_node.objects.filter(ind_deleted = False, type__in= db_type)
    all_cluster_ids = c123_node.objects.values_list("cluster__id", flat=True).distinct()
    licensing_queries = []
    no_touch_licensing = []
    current_trial_start = timezone.now() - relativedelta(months=2)
    license_db_type(targets, db_type, current_trial_start, all_cluster_ids, licensing_queries, no_touch_licensing, summary_count)
    licensing = c123_node.objects.none().union(*licensing_queries).values_list("id", flat=True)
    c123_node.objects.filter(id__in=licensing).update(ind_license = True)

def license_db_type(targets, db_type, current_trial_start, all_cluster_ids, licensing_queries, no_touch_licensing, licensing_summary):
    for cluster_id in all_cluster_ids:
        tmp_summary_count = 0 if licensing_summary < 0 else licensing_summary
        tmp_nodes_general_count = c123_node.objects.filter(type__in=db_type, cluster__id=cluster_id).count()
        tmp_nodes_deleted_count = c123_node.objects.filter(type__in=db_type, ind_deleted=True, cluster__id=cluster_id).count()

        if tmp_nodes_general_count == tmp_nodes_deleted_count:
            continue

        cluster = targets.filter(type__in=db_type, cluster__id=cluster_id)
        if not cluster.exists():
            continue

        tmp_nodes_count = cluster.count()
        old_nodes_count = cluster.filter(ind_license=True).count()
        new_nodes_count = cluster.filter(ind_license=False, created_date__gte=current_trial_start).count()
        expired_nodes_count = tmp_nodes_count - old_nodes_count - new_nodes_count

        if old_nodes_count > tmp_summary_count:
            cluster.update(ind_license=False)
            continue
        elif old_nodes_count == tmp_summary_count <= old_nodes_count + new_nodes_count and expired_nodes_count == 0:
            no_touch_licensing.append(cluster)
        elif old_nodes_count < tmp_summary_count <= old_nodes_count + new_nodes_count and expired_nodes_count == 0:
            max_change_to_true = min(tmp_summary_count - old_nodes_count, new_nodes_count)
            no_touch_licensing.append(cluster)
            cluster_filtered = cluster.filter(ind_license=False).order_by("created_date")[:max_change_to_true]
            licensing_queries.append(cluster_filtered)
        elif old_nodes_count <= tmp_summary_count <= old_nodes_count + new_nodes_count and expired_nodes_count > 0:
            if expired_nodes_count <= tmp_summary_count - old_nodes_count:
                no_touch_licensing.append(cluster)
                cluster_filtered = cluster.filter(ind_license=False, created_date__lt=current_trial_start).order_by("created_date")[:expired_nodes_count]
                licensing_queries.append(cluster_filtered)
                tmp_summary_count -= expired_nodes_count
                cluster_filtered = cluster.filter(ind_license=False, created_date__gte=current_trial_start).order_by("created_date")[:tmp_summary_count]
                licensing_queries.append(cluster_filtered)
                tmp_summary_count -= new_nodes_count
            else:
                cluster.update(ind_license=False)
            continue
        elif old_nodes_count <= tmp_summary_count and expired_nodes_count == 0:
            if new_nodes_count == 0 or new_nodes_count > tmp_summary_count:
                no_touch_licensing.append(cluster)
            elif new_nodes_count <= tmp_summary_count:
                licensing_queries.append(cluster)
        elif old_nodes_count <= tmp_summary_count and expired_nodes_count > 0:
            if old_nodes_count == tmp_summary_count:
                cluster.update(ind_license=False)
                continue
            else:
                if expired_nodes_count <= tmp_summary_count - old_nodes_count:
                    licensing_queries.append(cluster)
                else:
                    cluster.update(ind_license=False)
        else:
            licensing_queries.append(cluster)

        tmp_summary_count -= tmp_nodes_count

def allow_or_403(node):
    valid, expired_nodes = is_valid_job_target(node)
    if not valid:
        expired_names = ',\n'.join(expired_nodes.values_list("name", flat=True))
        error = f"Cluster of {node.name} has licensing issue.\nNodes with expired license:\n{expired_names}"
        raise APIException403(error)

def one_day_left(target_date):
    return target_date.day == (timezone.now()).day + 1

def is_valid_job_target(node):
    unlicensed_nodes = c123_node.objects.filter(type = node.type, cluster = node.cluster, ind_deleted = False, ind_license = False).exclude(type = DBType.HAPROXY.value)
    unlicensed_exists = unlicensed_nodes.exists()
    now = timezone.now()
    current_trial_start = now - relativedelta(months=2)
    expired_licenses = unlicensed_nodes.filter(created_date__lt=current_trial_start)
    expired_exists = expired_licenses.exists()
    valid = (now.month == 1 and now.day <= 14) or not expired_exists
    if unlicensed_exists:
        license_redemption_time = ''
        for i in unlicensed_nodes:
            # license notification email triggered only by notification-eligible nodes
            # check if we have such
            license_expiration_time = i.created_date + relativedelta(months=2)
            if one_day_left(license_expiration_time):
                license_redemption_time = 'in next one day'
                break
            if one_day_left(license_expiration_time - relativedelta(weeks=2)):
                license_redemption_time = "in next two weeks"
                break
            if one_day_left(license_expiration_time - relativedelta(weeks=1)):
                license_redemption_time = "in next one week"
                break

        if license_redemption_time:
            if expired_exists:
                license_redemption_time = "as soon as possible"
            nodes = '\n'.join(unlicensed_nodes.values_list("name", flat=True))
            user = node.created_by
            username = user.username
            email = user.email
            subject = "Cluster Licensing Notification"
            template = Template(unlicense_notice)
            context = Context({"user": username, "cluster": node.cluster.pk, "within_time": license_redemption_time, "nodes": nodes, "company": "123Cluster"})
            message = template.render(context)
            send_mail(subject, message, [email])
            logging.info(f'mail has sent to {email}')

    return (valid, expired_licenses)

def convert_char(char):
    mapping = {
        '0': '00000', '1': '00001', '2': '00002', '3': '00003', '4': '00004',
        '5': '00005', '6': '00006', '7': '00007', '8': '00008', '9': '00009',
        'A': '20000', 'B': '02000', 'C': '00200', 'D': '30000', 'E': '03000',
        'F': '00300', 'G': '40000', 'H': '04000', 'I': '00400', 'J': '50000',
        'K': '05000', 'L': '00500', 'M': '60000', 'N': '06000', 'O': '00600',
        'P': '70000', 'Q': '07000', 'R': '00700', 'S': '00070', 'T': '80000',
        'U': '08000', 'V': '00800', 'W': '90000', 'X': '09000', 'Y': '00900',
        'Z': '00090'
    }
    return mapping.get(char, char)

def convert_string(input_str):
    return ''.join(convert_char(char) for char in input_str)

def check_license_available(type):
    summary = generate_license_summary()
    match type:
        case DBType.ORACLE_CDB.value | DBType.ORACLE_PDB.value:
            avaliable_licenses = summary.oracle
            targets_count = c123_node.objects.filter(type__in=[DBType.ORACLE_CDB.value, DBType.ORACLE_PDB.value], ind_deleted=False).count()
        case DBType.MARIADB.value:
            avaliable_licenses = summary.mariadb
            targets_count = c123_node.objects.filter(type=DBType.MARIADB.value, ind_deleted=False).count()
        case DBType.MYSQL.value:
            avaliable_licenses = summary.mysql
            targets_count = c123_node.objects.filter(type=DBType.MYSQL.value, ind_deleted=False).count()
        case DBType.ELASTIC.value:
            avaliable_licenses = summary.elasticsearch
            targets_count = c123_node.objects.filter(type=DBType.ELASTIC.value, ind_deleted=False).count()
        case DBType.MONGO.value:
            avaliable_licenses = summary.mongo
            targets_count = c123_node.objects.filter(type=DBType.MONGO.value, ind_deleted=False).count()
        case DBType.MSSQL.value:
            avaliable_licenses = summary.mssql
            targets_count = c123_node.objects.filter(type=DBType.MSSQL.value, ind_deleted=False).count()
        case DBType.POSTGRES.value:
            avaliable_licenses = summary.postgresql
            targets_count = c123_node.objects.filter(type=DBType.POSTGRES.value, ind_deleted=False).count()
        case DBType.REDIS.value:
            avaliable_licenses = summary.redis
            targets_count = c123_node.objects.filter(type=DBType.REDIS.value, ind_deleted=False).count()
        case _ :
            return False

    return  avaliable_licenses > targets_count

def revoke_license(source, target = None, cluster = False):
    # for the drop node, to not repeat same two lines everywhere
    if not target and not cluster:
         source.ind_license = False
         source.save()
    # if we need unlicense entire cluster, for service purposes
    elif not target and cluster:
        c123_node.objects.filter(cluster = source).update(ind_license = False)
    else:
        type = source.type
        summary = generate_license_summary()
        from clusterCore.api.licensing.views import LicenseType
        if type in LicenseType.ORACLE.value:
            type = LicenseType.ORACLE.value
            total_licenses = summary.oracle
        elif type in LicenseType.MARIADB.value:
            type = LicenseType.MARIADB.value
            total_licenses = summary.mariadb
        elif type in LicenseType.MYSQL.value:
            type = LicenseType.MYSQL.value
            total_licenses = summary.mysql
        elif type in LicenseType.MONGO.value:
            type = LicenseType.MONGO.value
            total_licenses = summary.mongo
        elif type in LicenseType.MSSQL.value:
            type = LicenseType.MSSQL.value
            total_licenses = summary.mssql
        elif type in LicenseType.POSTGRES.value:
            type = LicenseType.POSTGRES.value
            total_licenses = summary.postgresql
        elif type in LicenseType.REDIS.value:
            type = LicenseType.REDIS.value
            total_licenses = summary.redis
        elif type in LicenseType.ELASTIC.value:
            type = LicenseType.ELASTIC.value
            total_licenses = summary.elasticsearch
        else:
            # if user make mess,or app have something incomplete
            print("move_license: incompatible cluster type ",  type) # dont expose actual value to frontend,log somewhere instead
            raise APIException404("this cluster type is not supported yet")

        licensed_of_type = c123_node.objects.filter(ind__deleted = False, type__in = type, ind_license = True)
        if not licensed_of_type.exists():
            # assume user make mess,license to move should exist
            print("move_license: no licenses of type ", type)
            raise APIException403("no licenses exists for this cluster type")

        licensed_count = licensed_of_type.count()
        available_licenses = total_licenses - licensed_count
        if available_licenses < 0:
            # assume user make mess,could not issue a license
            raise APIException500("move_license:'avaliability_less_than_total' mismatch")

        source_licensed = c123_node.objects.filter(cluster = source, ind_deleted = False, ind_license = True)
        if not source_licensed.exists():
            # assume user make mess,license to move should exist
            raise APIException403(f"no licenses to move from cluster {source.name}")

        source_licenses = source_licensed.count()
        targets = c123_node.objects.filter(cluster = target, ind_deleted = False, ind_license = False)
        targets_count = targets.count()
        if source_licenses < targets_count:
            license_pool = source_licenses + available_licenses
            if license_pool < targets_count:
                #here detailed (what details?) or simple?
                raise APIException403(f"Avaliable license pool are not enough to cover cluster {target.name}")

        source_licenses.update(ind_license = False)
        targets.update(ind_license = True)

def license_after_create(create):
    def wrapper(self, validated_data):
        node, task = create(self, validated_data)
        profile = Profile.objects.filter(user = node.created_by).first()
        if profile:
            if profile.is_demo:
                node.sizing = Node_Size_Allowed.SMALL.name
                node.auto_delete =  node.created_date + timezone.timedelta(days=1)
        if DBType.HAPROXY.value in node.type:
            # haproxy are not covered by our license key so available no matter what
            node.ind_license = True
            node.save()
        else:
            if profile:
                if profile.is_demo:
                    #not giving license to demo node
                    node.ind_license = False
                    node.save()
            else:
                available =  check_license_available(node.type)
                # redis has multiple equal masters, so primary needs additional check...
                if node.cluster_role == ClusterRoles.PRIMARY.value and (node.type != DBType.REDIS.value or (node.type == DBType.REDIS.value and not c123_node.objects.filter(ind_deleted = False, type = node.type, cluster = node.cluster, cluster_role = ClusterRoles.PRIMARY.value).exclude(pk = node.pk).exists())):
                    node.ind_license = available
                # not primary/not first primary
                else:
                    current_trial_start = timezone.now() - relativedelta(months=2)
                    expired_nodes_exists = c123_node.objects.filter(ind_deleted = False, ind_license = False, type = node.type, cluster = node.cluster, created_date__lt = current_trial_start).exists()
                    node.ind_license = available and not expired_nodes_exists
                node.save()

        return task

    return wrapper


def is_valid_uuid4(uuid_str):
    try:
        return UUID(uuid_str, version=4).version == 4
    except ValueError:
        return False

def ensure_primary_created(primary):
    match primary.type:
        case DBType.ELASTIC.value:
            task_name = 'install_elasticsearch_cluster_task'
        case DBType.MARIADB.value:
                task_name = 'install_mariadb_standalone_task'
        case DBType.MYSQL.value:
                task_name = 'install_mysql_standalone_task'
        case DBType.MONGO.value:
            task_name = 'install_mongo_node_task'
        case DBType.POSTGRES.value:
            task_name = 'install_postgres_standalone_task'
        # if used for some node where no need to maintain order of adding nodes (mssql and oracle, redis are in question)
        case _:
            return
    primary_creation_tasks = Task.objects.filter(task = task_name, target_id = primary.id)
    if primary_creation_tasks.filter(status__in = [TaskState.PENDING.value, TaskState.RUNNING.value]).exists():
        raise APIException409(PRIMARY_CREATING.format(primary.name))
    last_completed_creation = primary_creation_tasks.filter(status = TaskState.ON_DEMAND.value).last()
    if last_completed_creation:
        last_log = TaskLog.objects.filter(task_id = last_completed_creation.task_id).last()
        if last_log:
            if last_log.status != Severity.OK.value:
                raise APIException403(PRIMARY_CREATION_FAILED.format(primary.name))

def parse_license_key(key, update = True):
    try:
        decrypted = reverse_convert_string(key.strip())
    except Exception as e:
        raise APIException500(str(e))

    is_valid = check_pattern(decrypted, update)
    if not is_valid:
        raise APIException400("License key is invalid")

    sections = parse_sections(decrypted)
    # common fields of create/update
    model_dict ={
        'oracle': sections['ORC'],
        'mongo' : sections['MON'],
        'postgresql': sections['POS'],
        'mariadb' : sections['MDB'],
        'mysql' : sections['MYS'],
        'redis' : sections['RED'],
        'mssql' : sections['SQL'],
        'elasticsearch' : sections['ELS']
    }
    if not update:
        pk = decrypted.split('EXP')[0]
        if is_valid_uuid4(pk):
            pk = convert_string(pk)
            model_dict['uuid'] = pk
            model_dict['key'] = key.split(pk)[1]

    return model_dict

def reverse_convert_string(converted_str):
    reverse_mapping = {
        '00000': '0', '00001': '1', '00002': '2', '00003': '3', '00004': '4',
        '00005': '5', '00006': '6', '00007': '7', '00008': '8', '00009': '9',
        '20000': 'A', '02000': 'B', '00200': 'C', '30000': 'D', '03000': 'E',
        '00300': 'F', '40000': 'G', '04000': 'H', '00400': 'I', '50000': 'J',
        '05000': 'K', '00500': 'L', '60000': 'M', '06000': 'N', '00600': 'O',
        '70000': 'P', '07000': 'Q', '00700': 'R', '00070': 'S', '80000': 'T',
        '08000': 'U', '00800': 'V', '90000': 'W', '09000': 'X', '00900': 'Y',
        '00090': 'Z'
    }
    # Split the converted string into groups of 5 characters and join using the mapping
    return ''.join(reverse_mapping.get(segment, segment) for segment in [converted_str[i:i + 5] for i in range(0, len(converted_str), 5)])

def check_pattern(input_str, update):
    pattern = r'^EXP\d{8}ORC\d{3}MON\d{3}POS\d{3}MDB\d{3}MYS\d{3}RED\d{3}SQL\d{3}ELS\d{3}$'if update else r'^[A-F\d]{32}EXP\d{8}ORC\d{3}MON\d{3}POS\d{3}MDB\d{3}MYS\d{3}RED\d{3}SQL\d{3}ELS\d{3}$'
    return bool(re.match(pattern, input_str))

def parse_sections(string):
    pattern = r'(EXP|ORC|MON|POS|MDB|MYS|RED|SQL|ELS)(\d+)'  # Matches the designated sections
    matches = re.findall(pattern, string)
    return {
        key: 0 if value == '000'else int(value.lstrip('0'))
        for key, value in matches
    }

def generate_license_summary():
    summary, created = CustomerKey.objects.get_or_create(key = 'summary')
    distinct_keys = CustomerKey.objects.exclude(key = 'summary').values_list('key', flat=True).distinct()
    for key in distinct_keys :
        CustomerKey.objects.filter(key = key).update(**parse_license_key(key))

    other_keys = CustomerKey.objects.exclude(key = 'summary')
    other_keys_exist = other_keys.exists()
    aggregation_fields = {
        'elasticsearch': 'elasticsearch__sum',
        'mariadb': 'mariadb__sum',
        'mysql': 'mysql__sum',
        'mongo': 'mongo__sum',
        'mssql': 'mssql__sum',
        'oracle': 'oracle__sum',
        'postgresql': 'postgresql__sum',
        'redis': 'redis__sum',
    }
    for field, aggregation_key in aggregation_fields.items():
        setattr(summary, field, other_keys.aggregate(Sum(field))[aggregation_key] if other_keys_exist else 0)

    summary.save()
    return summary

def generate_random_password(length=12):
    """
    Generate a random complex password with uppercase, lowercase, digits, and symbols.
    :param length: Length of the password (default is 12)
    :return: Random complex password
    """
    import random
    import string
    
    characters = string.ascii_letters + string.digits + string.punctuation
    return ''.join(random.choices(characters, k=length))

def generate_human_readable_username():
    """
    Generate a human-readable random username.
    :return: Random username in the format "adjective_noun123"
    """
    import random
    
    adjectives = ["brave", "swift", "clever", "mighty", "fierce", "jolly", "sly", "nimble", "witty", "bold"]
    nouns = ["tiger", "eagle", "fox", "panther", "wolf", "otter", "hawk", "lynx", "bear", "falcon"]
    
    adjective = random.choice(adjectives)
    noun = random.choice(nouns)
    number = random.randint(100, 999)
    return f"{adjective}_{noun}{number}"

def fetch_data(query, cursor, params=None):
    cursor.execute(query, params)
    while True:
        rows = cursor.fetchmany(size=1000)
        if not rows:
            break
        for row in rows:
            yield row

def generate_connection_string(type, host, port, username, database):
    masked_password = "xxx"
    match type:
        case DBType.MONGO.value:
            return f"mongodb://{username}:{masked_password}@{host}:{port}/{database}"
        case DBType.REDIS.value:
            return f"redis://{username}:{masked_password}@{host}:{port}/0"
        case DBType.MYSQL.value | DBType.MARIADB.value:
            return f"mysql://{username}:{masked_password}@{host}:{port}/{database}"
        case DBType.POSTGRES.value:
            return f"postgresql://{username}:{masked_password}@{host}:{port}/{database}"
        case DBType.ELASTIC.value:
            return f"http://{username}:{masked_password}@{host}:{port}"
        case _:
            return ""

def format_rsa_key(key_string):
    if not key_string.strip():
        return key_string

    key_type = "RSA" if "RSA" in key_string else "OPENSSH"
    header = f"-----BEGIN {key_type} PRIVATE KEY-----"
    footer = f"-----END {key_type} PRIVATE KEY-----"
    # 80 for ED25519 PEM
    step = 64 if  key_type == "RSA" else 80
    if header not in key_string or footer not in key_string:
        raise APIException406("Not a valid PEM private key")

    # Split the key string into header, main part, and footer
    parts = key_string.split(header)
    main_part = parts[1].split(footer)[0]
    main_part = "".join(main_part.split())
    # Insert line breaks every 'step' characters in the main part
    main_part =  "\n".join(main_part[i:i+step] for i in range(0, len(main_part), step))
    # Combine the modified parts
    formatted_key =f"{header}\n{main_part}\n{footer}"
    return formatted_key

def cleanup_logs(obj_id, obj_type):
    try:
        with connection.cursor() as cur:
            cur.execute('DELETE FROM app_123monitor.rp_monitors WHERE node_id=%s AND db_type=%s', [obj_id, obj_type])
            cur.execute('DELETE FROM "clusterCore_task" WHERE target_type=%s AND target_id=%s', [obj_type,obj_id])
            cur.execute('DELETE FROM "clusterCore_tasklog" WHERE task_id=ANY(SELECT task_id FROM "clusterCore_task" WHERE target_type=%s AND target_id=%s)', [obj_type,obj_id])
            cur.execute('DELETE FROM "clusterCore_tasksteplog" WHERE task_log_id=ANY(SELECT task_log_id FROM "clusterCore_tasklog" WHERE task_id=ANY(SELECT task_id FROM "clusterCore_task" WHERE target_type=%s AND target_id=%s))', [obj_type,obj_id])
            connection.commit()
    except Exception as e:
        logging.error(cur.query)
        logging.error(e.strerror if hasattr(e, 'strerror') else str(e))

def find_replicaset_primary(node) -> str:


    try:
        client = create_mongo_connection(node)
        if isinstance(client, tuple):
            return 'none'

        info = client.admin.command('hello')
    except Exception as e:
        logging.error(f"find_replicaset_primary: {str(e.args[0])}")
        node.status = Status.DOWN.value
        node.severity = Severity.NOT_OK.value
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        node.ind_started = True
        node.save()
        return 'none'

    node.severity = Severity.OK.value
    node.status = Status.OK.value
    node.ind_started = True
    node.save()
    c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)

    rs = info.get('setName')
    if client.is_primary and not rs:
        return 'standalone'
    elif rs:
        primary = info.get('primary')
        if primary:
            return primary

    return 'none'

def find_replicaset_secondary(node, search) -> str:


    try:

        client = create_mongo_connection(node)
        if isinstance(client, tuple):
            return 'none'

        status = client.admin.command("replSetGetStatus")

        if search:
            found = next((i["name"].strip() for i in status["members"] if i["stateStr"].lower() == "secondary" and i["name"].strip() == search), None)
            if found:
                return found

        other = next((i["name"] for i in status["members"] if i["stateStr"].lower() == "secondary"), None)
        if other:
            return other

        return 'none'
    except Exception as e:
        node.status = Status.DOWN.value
        node.severity = Severity.NOT_OK.value
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        node.ind_started = True
        node.save()
        logging.error(f"find_replicaset_secondary: {str(e.args[0])}")
        return 'none'


def update_mongo_roles(app_primary):

    if app_primary.status in [Status.STOPPING.value, Status.STOPPED.value]:
        return

    should_skip = Task.objects.filter(task__in = ['mongo_backup_restore_task', 'restart_mongo_node_task', 'start_mongo_node_task', 'stop_mongo_node_task', 'stepdown_mongo_primary_task', 'delete_mongo_replicaset_node_task', 'delete_mongo_standalone_node_task'], target_id = app_primary.pk, status__in = [TaskState.PENDING.value, TaskState.RUNNING.value]).exists()

    if should_skip:
        return

    cluster = app_primary.cluster
    if not cluster.ex_params:
        cluster.ex_params = {'primary_count': 1}
        cluster.save()

    master = find_replicaset_primary(app_primary)
    if master == app_primary.name:
        cluster.ex_params['primary_count'] = 1
        cluster.save()

    if master != app_primary.name:
        standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MONGO.value, cluster = app_primary.cluster, ex_params__replicaset_name = app_primary.ex_params.get('replicaset_name'), ind_deleted = False)
        if not standbys.exists():
            cluster.ex_params['primary_count'] = 0
            cluster.save()
        else:
            #searching for new primary in primary standbys
            for standby in standbys:
                master = find_replicaset_primary(standby)
                if master == standby.name:
                    cluster.ex_params['primary_count'] = 1
                    cluster.save()
                    standby.cluster_role = ClusterRoles.PRIMARY.value
                    standby.ex_params["mongo_master_id"] = standby.id
                    standby.save()
                    app_primary.cluster_role = ClusterRoles.STANDBY.value
                    app_primary.ex_params["mongo_master_id"]
                    app_primary.save()
                    for other in standbys.exclude(pk = standby.pk):
                        other.cluster_role = ClusterRoles.STANDBY.value
                        other.ex_params["mongo_master_id"] = standby.id
                        other.save()
                    arbiters = c123_node.objects.filter(cluster_role = ClusterRoles.ARBITER.value, type = DBType.MONGO.value, cluster = app_primary.cluster, ex_params__replicaset_name = app_primary.ex_params.get('replicaset_name'), ind_deleted = False)
                    for arbiter in arbiters:
                        arbiter.ex_params["mongo_master_id"] = standby.id
                        arbiter.save()
                    break

            else:
                cluster.ex_params['primary_count'] = 0
                cluster.save()



    standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.MONGO.value, cluster = app_primary.cluster, ex_params__replicaset_name = app_primary.ex_params.get('replicaset_name'), ind_deleted = False).exclude(status = Status.DROPPING.value)
    for standby in standbys:
        standby.ex_params['lag'] = get_mongo_standby_lag(standby)
        standby.save()

def export_mongo_database(data):

    try:

        path = data.get('export_path')
        db_id = data.get('db_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        dump_name = data.get('dump_name')
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mongo_db = c123_database.objects.filter(name=db_id, node__name = node_id, type = DBType.MONGO.value, ind_deleted = False).first()
            else:
                mongo_db = c123_database.objects.filter(pk=db_id, type = DBType.MONGO.value, ind_deleted = False).first()
            if not mongo_db:
                raise APIException404("database id {} not found".format(db_id))

            if mongo_db.name == "local":
                raise APIException403("database local cant be dumped")

            if api_call:
                db_id = mongo_db.pk
            mongo_node = mongo_db.node
            allow_or_403(mongo_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MONGO.value)
            keep_days = getattr(size_scheme, f'{mongo_node.sizing.lower()}_export_keep_days', None)
            dump = serializer.create(serializer.validated_data)
            dump.name = dump_name
            dump.database = mongo_db
            dump.node = mongo_node
            if keep_days:
                dump.keep_days = keep_days
            dump.type = Inventory.EXPORT.value
            dump.created_by = created_by
            dump.ex_params = {"size": 0}
            dump.ex_params["belong_to"] = DBType.MONGO.value
            dump.ex_params["node_version"] = mongo_node.version
            dump.ex_params["duration"] = 0
            dump.ex_params["compress"] = compress
            dump.ex_params["modified_date"] = ''
            dump.ex_params["host_id"] = mongo_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            dump.name = "{}_{}".format(dump.name, now)
            dump.save()
            mongo_db.status = Status.BACKUP.value
            mongo_db.save()
            # serialize objects for the tasks
            db_name = mongo_db.name
            host_data = c123_HostSerializer(mongo_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            dump.ex_params["backup_path"] = f'{path}/dumps/{mongo_node.host.name}_{mongo_node.port}/{now}'
            dump.save()
            tmp_data = json.dumps({'db_id': db_id, 'db_name': db_name, 'node_data': object_to_json(mongo_node), 'dump_data': object_to_json(dump), 'host_data': host_data, 'backup_host': backup_host, 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='mongodump_task',
                                    displayment='dump mongo database {}'.format(mongo_db.name),
                                    target_type=DBType.MONGO.value,
                                    target_id=mongo_db.id,
                                    org_id = mongo_node.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=created_by)
            dump.ex_params['task_id'] = task.pk
            dump.save()
            return task
        else:
            raise APIException400(f"{serializer.errors}")

    except Exception as e:
        raise APIException500(str(e))

def export_mongo_node(data):

    try:

        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        path = data.get('export_path')
        dump_name = data.get('dump_name')
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")

        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mongo_node = c123_node.objects.filter(name=node_id, type = DBType.MONGO.value, ind_deleted = False).first()
            else:
                mongo_node = c123_node.objects.filter(pk=node_id, type = DBType.MONGO.value, ind_deleted = False).first()
            if not mongo_node:
                raise APIException406(f"node id {node_id} not available")
            allow_or_403(mongo_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MONGO.value)
            keep_days = getattr(size_scheme, f'{mongo_node.sizing.lower()}_export_keep_days', None)
            dump = serializer.create(serializer.validated_data)
            dump.name = dump_name
            dump.node = mongo_node
            if keep_days:
                dump.keep_days = keep_days
            dump.type = Inventory.EXPORT.value
            dump.created_by = created_by
            dump.ex_params = {"size": 0}
            dump.ex_params["belong_to"] = DBType.MONGO.value
            dump.ex_params["node_version"] = mongo_node.version
            dump.ex_params["duration"] = 0
            dump.ex_params["compress"] = compress
            dump.ex_params["modified_date"] = ''
            # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            dump.name = "{}_{}".format(dump.name, now)
            dump.save()
            mongo_node.status = Status.BACKUP.value
            mongo_node.save()
            # serialize objects for the tasks
            node_data = object_to_json(mongo_node)
            host_data = c123_HostSerializer(mongo_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            dump.ex_params["backup_path"] = f'{path}/dumps/{mongo_node.host.name}_{mongo_node.port}/{now}'
            dump.save()
            tmp_data = json.dumps({'db_name': None, 'node_data': node_data, 'host_data': host_data, 'backup_host': backup_host,  'dump_data': object_to_json(dump), 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='mongodump_task',
                                        displayment=f'dump mongo node {mongo_node.name}',
                                        target_type=DBType.MONGO.value,
                                        target_id=mongo_node.id,
                                        org_id = mongo_node.org_id,
                                        schedule=Schedule.NOW.value,
                                        status=TaskState.PENDING.value,
                                        parameters=tmp_data,
                                        created_by=created_by)
            dump.ex_params['task_id'] = task.pk
            dump.save()
            return task

        else:
            raise APIException400(f"{serializer.errors}")


    except Exception as e:
        raise APIException500(str(e))

def export_pg_database(data):

    try:

        path = data.get("export_path")
        db_id = data.get('db_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        include_publications = data.get('include_publications')
        include_subscriptions = data.get('include_subscriptions')
        use_inserts = data.get('use_inserts')
        plain = data.get('plain')
        export_name = data.get("export_name")
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                pg_db = c123_database.objects.filter(type = DBType.POSTGRES.value, name=db_id, node__name = node_id, ind_deleted = False).first()
            else:
                pg_db = c123_database.objects.filter(type = DBType.POSTGRES.value, pk=db_id, ind_deleted = False).first()
            if not pg_db:
                raise APIException410("database id {} not available".format(db_id))
            pg_node = pg_db.node
            allow_or_403(pg_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
            keep_days = getattr(size_scheme, f'{pg_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.database = pg_db
            export.node = pg_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.POSTGRES.value
            export.ex_params["node_version"] = pg_node.version
            export.ex_params["compress"] = compress
            export.ex_params["plain"] = plain
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params["host_id"] = pg_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export_name, now)
            # export.name = export_name
            export.save()
            pg_db.status = Status.BACKUP.value
            pg_db.save()
            # serialize objects for the tasks
            db_data = object_to_json(pg_db)
            host_data = c123_HostSerializer(pg_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            export.ex_params["compress"] = compress
            export.ex_params["backup_path"] = f'{path}/dumps/{pg_node.host.name}_{pg_node.port}/{now}'
            export.save()
            tmp_data = json.dumps({'db_data': db_data,
                                   'node_data': object_to_json(pg_node),
                                   'backup_data': object_to_json(export),
                                   'host_data': host_data,
                                   'backup_host':backup_host,
                                   'backup_type':'dump',
                                   'include_data': include_data,
                                   'include_publications': include_publications,
                                   'include_subscriptions': include_subscriptions,
                                   'use_inserts': use_inserts,
                                   'created_by': created_by.id})
            task = Task.objects.create(task='pg_dump_backup_task',
                                                displayment='backup {}'.format(pg_db.name),
                                                target_type=DBType.POSTGRES.value,
                                                target_id=pg_db.id,
                                                org_id=pg_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)
            export.ex_params['task_id'] = task.pk
            export.save()
            return task
        else:
            raise APIException400(f"{serializer.errors}")
    except Exception as e:
            raise APIException500(str(e))


def export_mysql_node(data):

    try:

        path = data.get('export_path')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        export_name = data.get("export_name")
        api_call = data.get("rest_api")
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")

        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mysql_node = c123_node.objects.filter(name=node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
            else:
                mysql_node = c123_node.objects.filter(pk=node_id, type = DBType.MYSQL.value, ind_deleted = False).first()

            if not mysql_node:
                raise APIException410("node id {} not available".format(node_id))
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MYSQL.value)
            keep_days = getattr(size_scheme, f'{mysql_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.node = mysql_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.MYSQL.value
            export.ex_params["compress"] = compress
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params["host_id"] = export.node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export_name, now)
            # export.name = export_name
            export.save()
            mysql_node.status = Status.BACKUP.value
            mysql_node.save()
            # serialize objects for the tasks
            host_data = c123_HostSerializer(mysql_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            export.ex_params["backup_path"] = f'{path}/{mysql_node.host.name}_{mysql_node.port}/{now}'
            export.save()
            tmp_data = json.dumps({'node_data': object_to_json(mysql_node), 'backup_data': object_to_json(export), 'host_data': host_data, 'backup_host': backup_host, 'backup_type':'dumpall', 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='mysqldump_task',
                                                displayment='backup {}'.format(mysql_node.name),
                                                target_type=DBType.MYSQL.value,
                                                target_id=mysql_node.id,
                                                org_id = mysql_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)
            export.ex_params['task_id'] = task.pk
            export.save()
            return task
        else:
            raise APIException400(f"{serializer.errors}")

    except Exception as e:
       raise APIException500(str(e))


def export_mssql_node(data):

    path = data.get("export_path")
    node_id = data.get('node_id')
    compress = data.get('compress')
    include_data = data.get('include_data')
    export_name = data.get('export_name')
    api_call = data.get('rest_api')
    created_by = User.objects.filter(id = data.get('created_by')).first()
    global SHELL_SAFE_TRANSFORM


    serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
    if serializer.is_valid():
        if api_call:
            mssql_node = c123_node.objects.filter(name=node_id, type = DBType.MSSQL.value, ind_deleted = False).first()
        else:
            mssql_node = c123_node.objects.filter(pk=node_id, type = DBType.MSSQL.value, ind_deleted = False).first()


        if not mssql_node:
            raise APIException410("node id {} not available".format(node_id))
        allow_or_403(mssql_node)
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")

        backup_host = object_to_json(app_host)
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MSSQL.value)
        keep_days = getattr(size_scheme, f'{mssql_node.sizing.lower()}_export_keep_days', None)
        export = serializer.create(serializer.validated_data)
        export.name = export_name
        export.node = mssql_node
        if keep_days:
            export.keep_days = keep_days
        export.type = Inventory.EXPORT.value
        export.created_by = created_by
        export.ex_params = {"size": 0}
        export.ex_params["belong_to"] = DBType.MSSQL.value
        export.ex_params["node_version"] = mssql_node.version
        export.ex_params["compress"] = compress
        export.ex_params["duration"] = 0
        export.ex_params["modified_date"] = ''
        export.ex_params['include_data'] = include_data
        # # create a unique string with the date
        now = timezone.now().strftime('%d%m%Y%H%M%S')
        export.name = "{}_{}".format(mssql_node.name, now).translate(SHELL_SAFE_TRANSFORM)
        # export.name = export_name
        export.save()
        mssql_node.status = Status.BACKUP.value
        mssql_node.save()
        # serialize objects for the tasks
        host_data = object_to_json(mssql_node.host)
        export.ex_params["backup_path"] = f'{path}/{mssql_node.host.name}_{mssql_node.port}/{now}'.replace(".", "_")
        export.save()
        tmp_data = json.dumps({'node_data': object_to_json(mssql_node), 'backup_data': object_to_json(export), 'host_data': host_data, 'backup_type':'dumpall', 'backup_host': backup_host, 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
        task = Task.objects.create(task='export_sqlserver_instance_task',
                                            displayment='backup {}'.format(mssql_node.name),
                                            target_type=DBType.MSSQL.value,
                                            target_id=mssql_node.id,
                                            org_id = mssql_node.org_id,
                                            schedule=Schedule.NOW.value,
                                            status=TaskState.PENDING.value,
                                            parameters=tmp_data,
                                            created_by=created_by)
        export.ex_params['task_id'] = task.pk
        export.save()
        return task
    else:
        print("invalid form")
        print(serializer.errors)
        return JsonResponse({'error': serializer.errors}, status=404)


def export_mariadb_node(data):

    try:

        path = data.get('export_path')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        export_name = data.get('export_name')
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mariadb_node = c123_node.objects.filter(name=node_id, type = DBType.MARIADB.value, ind_deleted = False).first()
            else:
                mariadb_node = c123_node.objects.filter(pk=node_id, type = DBType.MARIADB.value, ind_deleted = False).first()

            if not mariadb_node:
                raise APIException410("node id {} not available".format(node_id))
            allow_or_403(mariadb_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MARIADB.value)
            keep_days = getattr(size_scheme, f'{mariadb_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.node = mariadb_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.MYSQL.value
            export.ex_params["node_version"] = mariadb_node.version
            export.ex_params["compress"] = compress
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params["host_id"] = mariadb_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export.name, now)
            # export.name = export_name
            export.save()
            mariadb_node.status = Status.BACKUP.value
            mariadb_node.save()
            # serialize objects for the tasks
            host_data = c123_HostSerializer(mariadb_node.host).data
            export.ex_params["backup_path"] = f'{path}/{mariadb_node.host.name}_{mariadb_node.port}/{now}'
            export.save()
            backup_host = c123_HostSerializer(app_host).data
            tmp_data = json.dumps({'node_data': object_to_json(mariadb_node), 'backup_host': backup_host, 'backup_data': object_to_json(export), 'host_data': host_data, 'backup_type':'dumpall', 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='mysqldump_task',
                                                displayment='backup {}'.format(mariadb_node.name),
                                                target_type=DBType.MARIADB.value,
                                                target_id=mariadb_node.id,
                                                org_id = mariadb_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)
            export.ex_params['task_id'] = task.pk
            export.save()
            return task

        else:
            raise APIException400(f"{serializer.errors}")

    except Exception as e:
       raise APIException500(str(e))


def export_elasticsearch_index(data):

    try:

        index_id = data.get('index_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        export_path = data.get("export_path")
        dump_name = data.get('dump_name')
        api_call = data.get("rest_api")
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                elasticsearch_index = c123_database.objects.filter(name=index_id, node__name = node_id, type = DBType.ELASTIC.value, ind_deleted = False).first()
            else:
                elasticsearch_index = c123_database.objects.filter(pk=index_id, type = DBType.ELASTIC.value, ind_deleted = False).first()
            if not elasticsearch_index:
                raise APIException404("index id {} not found".format(index_id))

            elasticsearch_master = elasticsearch_index.node
            allow_or_403(elasticsearch_master)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.ELASTIC.value)
            keep_days = getattr(size_scheme, f'{elasticsearch_master.sizing.lower()}_export_keep_days', None)
            dump = serializer.create(serializer.validated_data)
            dump.name = dump_name
            dump.database = elasticsearch_index
            dump.node = elasticsearch_master
            dump.type = Inventory.EXPORT.value
            dump.created_by = created_by
            dump.ex_params = {"size": 0}
            if keep_days:
                dump.keep_days = keep_days
            dump.ex_params["belong_to"] = DBType.ELASTIC.value
            dump.ex_params["node_version"] = elasticsearch_master.version
            dump.ex_params["backup_type"] = "index_snapshot"
            dump.ex_params["duration"] = 0
            dump.ex_params["compress"] = compress
            dump.ex_params["modified_date"] = ''
            dump.ex_params["host_id"] = elasticsearch_master.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            dump.name = "{}_{}".format(dump.name, now)
            dump.save()
            elasticsearch_index.status = Status.BACKUP.value
            elasticsearch_index.save()
            # serialize objects for the tasks
            index_name = elasticsearch_index.name
            backup_host = c123_HostSerializer(app_host).data
            dump.ex_params["backup_path"] = export_path
            dump.save()
            cluster_nodes = c123_node.objects.filter(cluster = elasticsearch_master.cluster, type = DBType.ELASTIC.value, ind_deleted = False)
            cluster_data = [{"host": object_to_json(i.host), "node": object_to_json(i)} for i in cluster_nodes]
            tmp_data = json.dumps({'cluster_data': cluster_data, 'backup_host': backup_host, 'db_id': index_id, 'db_name': index_name, 'backup_data': object_to_json(dump), 'compress': compress, 'created_by': created_by.id})
            task = Task.objects.create(task='elasticsearch_snapshot_task',
                                    displayment='snapshot elasticsearch index {}'.format(elasticsearch_index.name),
                                    target_type=DBType.ELASTIC.value,
                                    target_id=elasticsearch_index.id,
                                    org_id = elasticsearch_master.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=created_by)
            dump.ex_params['task_id'] = task.pk
            dump.save()
            return task

        else:
            raise APIException400(f"{serializer.errors}")

    except Exception as e:
        raise APIException500(str(e))



def export_mariadb_database(data):

    try:

        path = data.get('export_path')
        db_id = data.get('db_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        export_name = data.get('export_name')
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mariadb_db = c123_database.objects.filter(type = DBType.MARIADB.value, name=db_id, node__name = node_id, ind_deleted = False).first()
            else:
                mariadb_db = c123_database.objects.filter(type = DBType.MARIADB.value, pk=db_id, ind_deleted = False).first()
            if not mariadb_db:
                raise APIException404("database id {} not found".format(db_id))

            mariadb_node = mariadb_db.node
            allow_or_403(mariadb_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MARIADB.value)
            keep_days = getattr(size_scheme, f'{mariadb_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.database = mariadb_db
            export.node = mariadb_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.MARIADB.value
            export.ex_params["node_version"] = mariadb_node.version
            export.ex_params["compress"] = compress
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params["host_id"] = mariadb_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export.name, now)
            # export.name = export_name
            export.save()
            mariadb_db.status = Status.BACKUP.value
            mariadb_db.save()
            # serialize objects for the tasks
            db_data = object_to_json(mariadb_db)
            host_data = c123_HostSerializer(mariadb_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            export.ex_params["backup_path"] = f'{path}/{mariadb_node.host.name}_{mariadb_node.port}/{now}'
            export.save()
            tmp_data = json.dumps({'db_data': db_data, 'node_data': object_to_json(mariadb_node), 'backup_data': object_to_json(export), 'backup_host': backup_host, 'host_data': host_data, 'backup_type':'dump', 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='mysqldump_task',
                                                displayment='backup {}'.format(mariadb_db.name),
                                                target_type=DBType.MARIADB.value,
                                                target_id=mariadb_db.id,
                                                org_id = mariadb_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)
            export.ex_params['task_id'] = task.pk
            export.save()
            return task
        else:
            raise APIException400(f"{serializer.errors}")

    except Exception as e:
       raise APIException500(str(e))

def export_cdb_pdb(data):

    try:

        pdb_id = data.get('pdb_id')
        cdb_id = data.get('cdb_id')
        obj_type = data.get('obj_type')
        schema = data.get('schema')[0]
        file_name = data.get('file_name')
        keep_days = data.get('keep_days')
        path = data.get('disk')[1]
        db_path = data.get('disk')[0]
        include_data = data.get('include_data')
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()

        if api_call:
            location = cdb_id.split('@')
            cdb = c123_node.objects.filter(name=location[0], host__name = location[1],  type = DBType.ORACLE_CDB.value,  ind_deleted = False).first()
        else:
            cdb = c123_node.objects.filter(pk=cdb_id, type = DBType.ORACLE_CDB.value, ind_deleted = False).first()
        if not cdb:
            raise APIException410(f"CDB ID {cdb_id} is not available")


        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")

        size_scheme = OracleClusterSizing.objects.first()
        backup_host = c123_HostSerializer(app_host).data
        # create the export model
        if ClusterRoles.CDB.value in obj_type:
            allow_or_403(cdb)
            cdb.status = Status.BACKUP.value
            cdb.save()

            ex_params = {}
            ex_params['schema'] = schema
            ex_params['dump_file'] = file_name,
            ex_params['log_file'] = file_name.replace('dmp', 'log'),
            ex_params['remote_path'] = path
            ex_params['db_path_name'] = db_path
            ex_params['size'] = 0
            ex_params['duration'] = 0
            ex_params['exit_code'] = 1
            ex_params['belong_to'] = DBType.ORACLE_CDB.value
            ex_params['node_version'] = cdb.version
            ex_params["modified_date"] = ''
            keep_days = getattr(size_scheme, f'{cdb.sizing.lower()}_export_keep_days', None)
            export = c123_inventory.objects.create(
                created_by=created_by,
                node=cdb,
                type=Inventory.EXPORT.value,
                ex_params=ex_params,
                keep_days=int(keep_days),
                name=file_name,
                )
            export_data = object_to_json(export)
            cdb_data = object_to_json(cdb)
            host_data = c123_HostSerializer(cdb.host).data
            tmp_data = json.dumps({'cdb_data': cdb_data, 'host_data': host_data, 'backup_host': backup_host, 'export': export_data, 'target_type': 'cdb', 'include_data': include_data, 'created_by': created_by.id})
        elif ClusterRoles.PDB.value in obj_type:
            if api_call:
                location = pdb_id.split('@')
                pdb = c123_node.objects.filter(name=location[0], host__name=location[1], port=location[2], type = DBType.ORACLE_PDB.value, ind_deleted = False).first()
            else:
                pdb = c123_node.objects.filter(pk=pdb_id,  type = DBType.ORACLE_PDB.value, ind_deleted = False).first()
            if not pdb:
                raise APIException410(f"PDB ID {pdb_id} is not available")
            allow_or_403(pdb)
            pdb.status = Status.BACKUP.value
            pdb.save()
            ex_params = {}
            ex_params['schema'] = schema
            ex_params['dump_file'] = file_name,
            ex_params['log_file'] = file_name.replace('dmp', 'log'),
            ex_params['remote_path'] = path
            ex_params['db_path_name'] = db_path
            ex_params['cdb_id'] = cdb.pk
            ex_params['size'] = 0
            ex_params['duration'] = 0
            ex_params['exit_code'] = 1
            ex_params['belong_to'] = DBType.ORACLE_PDB.value
            ex_params['node_version'] = pdb.version
            keep_days = getattr(size_scheme, f'{pdb.sizing.lower()}_export_keep_days', None)

            export = c123_inventory.objects.create(
                created_by=created_by,
                node=pdb,
                type=Inventory.EXPORT.value,
                ex_params=ex_params,
                keep_days=int(keep_days),
                name=file_name
            )
            export_data = object_to_json(export)
            cdb_data = object_to_json(cdb)
            pdb_data = object_to_json(pdb)
            host_data = c123_HostSerializer(pdb.host).data
            backup_host = c123_HostSerializer(app_host).data
            tmp_data = json.dumps({'cdb_data': cdb_data, 'pdb_data': pdb_data, 'host_data': host_data, 'backup_host': backup_host, 'export': export_data, 'target_type': 'pdb', 'include_data': include_data, 'created_by': created_by.id})

        if ClusterRoles.CDB.value in obj_type:
            task = Task.objects.create(task='export_cdb_pdb_task',
                                    displayment='export CDB {}'.format(cdb.name),
                                    target_type=DBType.ORACLE_CDB.value,
                                    target_id=cdb.id,
                                    org_id=cdb.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=created_by)
        else:
            task = Task.objects.create(task='export_cdb_pdb_task',
                                    displayment=f'export PDB {pdb.name}',
                                    target_type=DBType.ORACLE_PDB.value,
                                    target_id=pdb.id,
                                    org_id=pdb.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=created_by)

        export.ex_params['task_id'] = task.pk
        export.save()
        return task

    except Exception as e:
        raise APIException500(str(e))

def export_mssql_database(data):

    try:

        path = data.get("export_path")
        db_id = data.get('db_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        export_name = data.get('export_name')
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mssql_db = c123_database.objects.filter(name=db_id, node__name = node_id, type = DBType.MSSQL.value, ind_deleted = False).first()
            else:
                mssql_db = c123_database.objects.filter(pk=db_id, type = DBType.MSSQL.value, ind_deleted = False).first()
            if not mssql_db:
                raise APIException410("database id {} not available".format(db_id))

            mssql_node = mssql_db.node
            allow_or_403(mssql_node)
            app_host = c123_host.objects.filter(ex_params__myself = True).first()
            if not app_host:
                raise APIException404("could not determine the app host")

            backup_host = object_to_json(app_host)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MSSQL.value)
            keep_days = getattr(size_scheme, f'{mssql_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.database = mssql_db
            export.node = mssql_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.MSSQL.value
            export.ex_params["node_version"] = mssql_node.version
            export.ex_params["compress"] = compress
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params['include_data'] = include_data
            export.ex_params["host_id"] = mssql_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(mssql_db.name, now).translate(SHELL_SAFE_TRANSFORM)
            # export.name = export_name
            export.save()
            mssql_db.status = Status.BACKUP.value
            mssql_db.save()
            # serialize objects for the tasks
            db_data = object_to_json(mssql_db)
            host_data = object_to_json(mssql_node.host)
            export.ex_params["backup_path"] = f'{path}/{mssql_node.host.name}_{mssql_node.port}/{now}'.replace(".", "_")
            export.save()
            tmp_data = json.dumps({'db_data': db_data, 'node_data': object_to_json(mssql_node), 'backup_data': object_to_json(export), 'host_data': host_data, 'backup_type':'dump', 'backup_host': backup_host, 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='export_sqlserver_database_task',
                                                displayment='backup {}'.format(mssql_db.name),
                                                target_type=DBType.MSSQL.value,
                                                target_id=mssql_db.id,
                                                org_id = mssql_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)

            export.ex_params['task_id'] = task.pk
            export.save()
            return task
        else:
           raise APIException400(f"{serializer.errors}")


    except Exception as e:
       raise APIException500(str(e))

def export_mysql_database(data):

    try:

        path = data.get('export_path')
        db_id = data.get('db_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        export_name = data.get("export_name")
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                mysql_db = c123_database.objects.filter(name=db_id, node__name = node_id, type = DBType.MYSQL.value, ind_deleted = False).first()
            else:
                mysql_db = c123_database.objects.filter(pk=db_id, type = DBType.MYSQL.value, ind_deleted = False).first()

            if not mysql_db:
                raise APIException410("database id {} not available".format(db_id))
            mysql_node = mysql_db.node
            allow_or_403(mysql_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MYSQL.value)
            keep_days = getattr(size_scheme, f'{mysql_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.database = mysql_db
            export.node = mysql_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.MYSQL.value
            export.ex_params["node_version"] = mysql_node.version
            export.ex_params["compress"] = compress
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params["host_id"] = mysql_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export.name, now)
            # export.name = export_name
            export.save()
            mysql_db.status = Status.BACKUP.value
            mysql_db.save()
            # serialize objects for the tasks
            db_data = object_to_json(mysql_db)
            host_data = c123_HostSerializer(mysql_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            export.ex_params["backup_path"] = f'{path}/{mysql_node.host.name}_{mysql_node.port}/{now}'
            export.save()
            tmp_data = json.dumps({'db_data': db_data, 'node_data': object_to_json(mysql_node), 'backup_data': object_to_json(export), 'host_data': host_data, 'backup_host': backup_host,  'backup_type':'dump', 'compress': compress, 'include_data': include_data, 'created_by': created_by.id})
            task = Task.objects.create(task='mysqldump_task',
                                                displayment='backup {}'.format(mysql_db.name),
                                                target_type=DBType.MYSQL.value,
                                                target_id=mysql_db.id,
                                                org_id = mysql_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)
            export.ex_params['task_id'] = task.pk
            export.save()
            return task
        else:
           raise APIException400(f"{serializer.errors}")

    except Exception as e:
       raise APIException500(str(e))

def export_pg_node(data):

    try:
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        include_publications = data.get('include_publications')
        include_subscriptions = data.get('include_subscriptions')
        use_inserts = data.get('use_inserts')
        path = data.get("export_path")
        export_name = data.get("export_name")
        api_call = data.get("rest_api")
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                pg_node = c123_node.objects.filter(type = DBType.POSTGRES.value, name=node_id, ind_deleted = False).first()
            else:
                pg_node = c123_node.objects.filter(type = DBType.POSTGRES.value, pk=node_id, ind_deleted = False).first()
            if not pg_node:
                raise APIException410("node id {} not available".format(node_id))
            allow_or_403(pg_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
            keep_days = getattr(size_scheme, f'{pg_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.node = pg_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.POSTGRES.value
            export.ex_params["node_version"] = pg_node.version
            export.ex_params["backup_type"] = "dumpall"
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export.name, now)
            export.save()
            pg_node.status = Status.BACKUP.value
            pg_node.save()
            # serialize objects for the tasks
            node_data = object_to_json(pg_node)
            host_data = c123_HostSerializer(pg_node.host).data
            export.ex_params["backup_path"] = f'{path}/dumps/{pg_node.host.name}_{pg_node.port}/{now}'
            export.ex_params["compress"] = compress
            export.save()
            backup_host = c123_HostSerializer(app_host).data
            tmp_data = json.dumps({'node_data': node_data,
                                   'host_data': host_data,
                                   'backup_host': backup_host,
                                   'backup_data': object_to_json(export),
                                   'backup_type':'dumpall',
                                   'include_data': include_data,
                                   'include_publications': include_publications,
                                   'include_subscriptions': include_subscriptions,
                                   'use_inserts': use_inserts,
                                   'created_by': created_by.id})
            task = Task.objects.create(task='pg_dump_backup_task',
                                        displayment='backup pg {}'.format(pg_node.name),
                                        target_type=DBType.POSTGRES.value,
                                        target_id=pg_node.id,
                                        org_id=pg_node.org_id,
                                        schedule=Schedule.NOW.value,
                                        status=TaskState.PENDING.value,
                                        parameters=tmp_data,
                                        created_by=created_by)

            export.ex_params['task_id'] = task.pk
            export.save()
            return task

        else:
            raise APIException400(f"{serializer.errors}")
    except Exception as e:
        raise APIException500(str(e))


def check_host_connection(node):
    try:
        host = node.host
        if DBType.MSSQL.value in node.type:
            client = create_winrm_connection(host.name, host.username, host.password, host.ssh_port)
        else:
            client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)

        if isinstance(client, tuple):
            node.severity = Severity.NOT_OK.value
            node.status = Status.OFFLINE.value
            node.save()

        return client
    except Exception as e:
        logging.error(f"check_host_connection: {e}")
        return (False, e)

def get_postgres_connection(node, db_name="postgres"):
    try:

        result = check_host_connection(node)

        global TASK_CERTIFICATE_FILES_DIR
        conn_params = {
                'dbname': db_name,
                'user':  node.database_user,
                'password':  node.database_pwd,
                'host': node.host.name.strip(),
                'port': node.port,
                'connect_timeout': Timeout.BACKEND_POSTG.value
        }
        certificate_authentication = node.ex_params.get('certificate_authentication')
        if certificate_authentication:
            certificate_root = f"{TASK_CERTIFICATE_FILES_DIR}/{conn_params['host']}_{conn_params['port']}"
            conn_params['sslmode'] = 'verify-full'
            conn_params['sslrootcert'] = f"{certificate_root}/root.crt"
            conn_params['sslcert'] = f"{certificate_root}/client.crt"
            conn_params['sslkey'] = f"{certificate_root}/client.key"
        conn = psycopg2.connect(**conn_params)
        node.severity = Severity.OK.value
        node.status = Status.OK.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        return conn
    except Exception as e:
        if not isinstance(result, tuple):
            if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                node.status = Status.STOPPED.value
            else:
                node.status = Status.DOWN.value

            node.severity = Severity.NOT_OK.value
            node.ind_started = False
            node.save()

            c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        logging.error(f"get_postgres_connection: {e}")
        return (False, e)


def connect_to_oracle_database_using_params(**kwargs):
    """return a oracledb connection object
    kwargs contain the host, port, sname, user, pwd to create
    the connection"""

    try:
        user = kwargs.pop('user')
        pwd = kwargs.pop('pwd')
        host = kwargs.pop('host')
        port = kwargs.pop('port')
        sname = kwargs.pop('sname')
        dsn_tns = oracledb.makedsn(host, port, sname)
        conn = oracledb.connect(user=user, password=pwd, dsn=dsn_tns, mode=oracledb.SYSDBA, tcp_connect_timeout=Timeout.BACKEND_ORACLE.value)
    except oracledb.DatabaseError as e:
        error, = e.args
        print(error)
        raise APIException401(str(error))
    else:
        return conn

def connect_to_sqlserver_instance(node, db = 'master', dry_run = False):
    """connecting to sql server"""

    try:
        result = check_host_connection(node)
        server = f"{node.host.name},{str(node.prt)}"
        drivers = pyodbc.drivers()
        if not drivers:
            return(False, "No ODBC drivers found on this system")
        else:
            versions = [int(driver.split()[2]) for driver in drivers if driver.startswith('ODBC Driver')]
            if not versions:
                return(False, "No ODBC drivers found for SQL Server on this system")
            else:
                newest_driver = f"ODBC Driver {max(versions)} for SQL Server"

        if dry_run:
            dsn = f'DRIVER={{{newest_driver}}};SERVER={server};DATABASE={db};UID={node.database_user};PWD=xxx;TrustServerCertificate=yes;'
            return dsn
        else:
            dsn = f'DRIVER={{{newest_driver}}};SERVER={server};DATABASE={db};UID={node.database_user};PWD={node.database_pwd};TrustServerCertificate=yes;'

        conn = pyodbc.connect(dsn, timeout=Timeout.SERVICE_MSSQL.value)

        return conn
    except Exception as e:
        if isinstance(node, c123_node):
            if not isinstance(result, tuple):
                if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                    node.status = Status.STOPPED.value
                else:
                    node.status = Status.DOWN.value

                node.severity = Severity.NOT_OK.value
                node.ind_started = False
                node.save()
                c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)

            logging.error(f"connect_to_sqlserver_instance: {e}")
            return (False, e)


def find_postgres_bin(client, postgres_version) -> str:
    """find location of POSTGRES BIN"""

    jsn = json.loads(remote_command(client, f'dirname `rpm -ql postgresql{postgres_version}-server.x86_64|grep initdb|grep -v share`', True), strict=False)
    if jsn['stdout'] != '':
        postgres_bin = jsn['stdout'].strip()

    if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
        if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
            postgres_bin = f'/usr/lib/postgresql/{postgres_version}/bin'

        else:
            raise Exception('find_postgres_bin:', jsn['stderr']) #???

    return postgres_bin

def find_elastic_bin(client) -> str:
    """find location of elasticsearch bin folder"""

    elastic_bin = '/usr/share/elasticsearch/bin'
    jsn = json.loads(remote_command(client, f"sudo rpm -ql elasticsearch | grep '/bin$' | grep -vE '(jdk|modules)'", True), strict=False)
    if jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower() and 'permission denied' not in jsn['stderr'].lower():
        if 'not found' in jsn['stderr'].lower():
            jsn = json.loads(remote_command(client, f"sudo rpm -ql elasticsearch | grep '/bin$' | grep -vE '(jdk|modules)'", True), strict=False)
            if jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower() and 'permission denied' not in jsn['stderr'].lower():
                raise Exception('find_elastic_bin:', jsn['stderr'])
        else:
            raise Exception('find_elastic_bin:', jsn['stderr'])

    if jsn['stdout'] != '':
        elastic_bin = jsn['stdout'].strip()

    return elastic_bin

def find_redis_bin(client) -> str:
    """find location of redis bin folder"""

    redis_bin = ''
    jsn = json.loads(remote_command(client, "dirname $(which redis-server)", True), strict=False)
    if jsn['stdout'] != '':
        redis_bin = jsn['stdout'].strip()

    err_lower = jsn['stderr'].lower()
    if err_lower != '' and not any (err in err_lower for err in ["warning", 'permission denied']):
        if any(err in err_lower for err in ['not found', 'missing operand']):
            redis_bin = f'/usr/bin'

        else:
            raise Exception('find_redis_bin:', jsn['stderr'])

    return redis_bin

def create_elasticsearch_insecure_connection(node):
    try:
        result = check_host_connection(node)
        client = Elasticsearch(f"https://{node.host}:{node.port}", basic_auth=(node.database_user, node.database_pwd), verify_certs= False, timeout=Timeout.BACKEND_ELASTIC.value)
        client.info()
        node.severity = Severity.OK.value
        node.status = Status.OK.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        return client
    except Exception as e:
        if not isinstance(result,tuple):
            if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                node.status = Status.STOPPED.value
            else:
                node.status = Status.DOWN.value

            node.severity = Severity.NOT_OK.value
            node.ind_started = False
            node.save()
            c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        logging.error(f"create_elasticsearch_insecure_connection: {e}")
        return (False, e)

def create_mongo_connection(node):
    try:
        result = check_host_connection(node)
        use_ssl = node.ex_params.get('ssl')
        kwargs = {
            'username': node.database_user,
            'password': node.database_pwd,
            'readPreference': 'primaryPreferred',
            'directConnection': True,
            'serverSelectionTimeoutMS': Timeout.BACKEND_MONGO.value
        }
        if use_ssl:
            cert_root = f"{TASK_CERTIFICATE_FILES_DIR}/{node.host.name}_{node.port}"
            ca = f"{cert_root}/ca.crt"
            key = f"{cert_root}/ca.pem"
            kwargs['tls'] = True
            kwargs['tlsCAFile'] = ca
            kwargs['tlsCertificateKeyFile'] = key
            kwargs['tlsAllowInvalidHostnames'] = True


        client = MongoClient(node.host.name, int(node.ex_params.get('mongos_port') if ClusterRoles.BALANCER.value in node.cluster_role else node.port), **kwargs)
        if ClusterRoles.BALANCER.value not in node.cluster_role:
            status = client.admin.command("replSetGetStatus")
            state_str = next((i["stateStr"] for i in status["members"] if i["name"].strip() == node.name), None)
            node.status = str(state_str)
        node.severity = Severity.OK.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        return client
    except Exception as e:
        if not isinstance(result, tuple):
            if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                node.status = Status.STOPPED.value
            else:
                node.status = Status.DOWN.value

            node.severity = Severity.NOT_OK.value
            node.ind_started = False
            node.save()
            c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        logging.error(f"create_mongo_connection: {e}")
        return (False, e)

def update_oracle_roles(cdb):
    try:
        if cdb.status in [Status.STOPPING.value, Status.STOPPED.value]:
            return
        cluster_nodes = c123_node.objects.filter(cluster = cdb.cluster, ind_deleted = False)
        conn = connect_to_oracle_database(cdb = cdb)
        if(isinstance(conn, dict)): #ora-1252 = cdb does not exist
            print("update_oracle_cluster_roles:", cdb.name, str(conn["message"]))
            cluster_nodes.update(severity = Severity.NOT_OK.value, status = Status.FAILED.value, ind_started = False)
            return
        cursor = conn.cursor()
        pdbs = cluster_nodes.filter(cluster_role = ClusterRoles.PDB.value)
        for pdb in pdbs:
            try:
                cursor.execute(f"ALTER SESSION SET CONTAINER={pdb.name}")
                pdb.severity = Severity.OK.value
                cursor.execute("SELECT open_mode FROM v$database")
                res = cursor.fetchall()
                if res:
                    if 'mounted' in res[0][0].lower():
                        pdb.ind_started = False
                        pdb.status = Status.CLOSED.value
                    else:
                        pdb.ind_started = True
                        pdb.status = Status.OK.value
                pdb.save()
            except Exception as pdb_error:
                print("update_oracle_cluster_roles:", cdb.name, pdb.name, str(pdb_error))
                pdb.severity = Severity.NOT_OK.value
                pdb.status = Status.DOWN.value
                pdb.ind_started = False
                pdb.save()
    except Exception as error:
        #here comes the other cdb-related exceptions
        print("update_oracle_cluster_roles:", cdb.name, str(error))
        cluster_nodes.update(severity = Severity.NOT_OK.value, status = Status.DOWN.value)
        return

def check_if_mysql_fork_primary(node):



    try:
        conn = create_mysql_connection(node)
        if isinstance(conn, tuple):
            return False
        cur = conn.cursor()
        cur.execute(f"SHOW {'REPLICA' if DBType.MYSQL.value in node.type and float(node.version) > 5.7 else 'SLAVE'} STATUS;")
        res = cur.fetchone()
        if not res:
            return True
        return False
    except Exception as e:
        node.status = Status.DOWN.value
        node.severity = Severity.NOT_OK.value
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        error = e.strerror if hasattr(e, 'strerror') else str(e)
        print("check_if_mysql_fork_primary:", error)
        return False


def update_mysql_fork_roles(app_primary):
    if app_primary.status in [Status.STOPPING.value, Status.STOPPED.value]:
        return
    app_primary_host = app_primary.host
    type = app_primary.type.lower()
    use_ssl = app_primary.ex_params['ssl']
    mysql_version = float(app_primary.version)
    error_label = f'update_{type}_roles:'
    link_field = f'{type}_master_id'
    mysql_app = DBType.MYSQL.value.lower()
    mariadb_app = DBType.MARIADB.value.lower()
    (distributor, release) = check_os_version(app_primary_host.__dict__)
    if OperatingSystem.UNKNOWN.value in distributor:
        raise APIException401(release['stderr'])
    appendix = 'd' if type in mysql_app and OperatingSystem.CENTOS.value in distributor else ''
    conf_variable = 'read_only' if mariadb_app in type else 'super_read_only'
    if type in mariadb_app:
        conf_path = "/etc/my.cnf.d/server.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/mariadb.conf.d/50-server.cnf"
    else:
        conf_path = "/etc/my.cnf" if OperatingSystem.CENTOS.value in distributor else "/etc/mysql/my.cnf"

    if use_ssl:
        plugin_type = 'sha256_password' if mysql_version < 8.0 else 'caching_sha2_password'
    else:
        plugin_type = 'mysql_native_password'

    authentication_plugin = f"WITH {plugin_type} BY" if mysql_app in type else f'VIA {plugin_type} USING'
    if mariadb_app in type or (mysql_app in type and mysql_version < 8.0):
        slave_label = 'SLAVE'
        master_label = 'MASTER'
        master_type = ''
    else:
        slave_label = 'REPLICA'
        master_label = 'SOURCE'
        master_type = 'REPLICATION'


    should_stop = Task.objects.filter(task__in=[ f'restore_{type}_node_task',
                                                f'switchover_{type}_node_task',
                                                f'install_{type}_standalone_task',
                                                f'upgrade_{type}_node_task',
                                                f'drop_{type}_node_task'],
                                                status__in = [TaskState.PENDING.value, TaskState.RUNNING.value], target_id = app_primary.pk).exists()
    if should_stop:
        return

    master_client = create_mysql_connection(app_primary)
    if isinstance(master_client, tuple):
        slaves = c123_node.objects.filter(cluster = app_primary.cluster, cluster_role = ClusterRoles.STANDBY.value,  type = app_primary.type, severity = Severity.OK.value, ind_started = True,  ind_deleted = False)
        for slave in slaves:
            slave_host = slave.host
            slave_client = create_mysql_connection(slave)
            if isinstance(slave_client, tuple):
                continue

            ssh_client = create_ssh_connection(slave_host.name, slave_host.username, slave_host.password, slave_host.private_key, prt = slave_host.ssh_port)
            if isinstance(ssh_client, tuple):
                print(error_label, slave_host.name,ssh_client[1])
                slave.severity = Severity.NOT_OK.value
                slave.status = Status.OFFLINE.value
                slave.save()
                continue

            slave.status = 'resetting replication'
            slave.severity = Severity.OK.value
            slave.ind_started = True
            slave.save()
            slave_reset_error = False
            commands = [
                f"sudo sed -i '/{conf_variable}=ON/d' '{conf_path}'",
                f'sudo systemctl restart {type}{appendix}',
            ]
            for command in commands:
                jsn = json.loads(remote_command(ssh_client, command, True), strict=False)
                if jsn['stderr'] != '' and any(err in jsn['stderr'].lower() for err in ['error', 'fail']):
                    print(error_label, slave.name, jsn['stderr'])
                    slave.status = 'failed to made read-write'
                    slave.severity = Severity.WARNING.value
                    slave.save()
                    slave_reset_error = True
                    break

            if slave_reset_error:
                continue

            slave_reset_error = False
            time.sleep(10)
            slave_client = create_mysql_connection(slave)
            if isinstance(slave_client, tuple):
                continue

            try:
                cur = slave_client.cursor()
                cur.execute(f"STOP {slave_label};")
                cur.execute(f"RESET {slave_label} ALL;")
            except Exception as e:
                error = e.strerror if hasattr(e, 'strerror') else str(e)
                print(error_label, slave.name,error)
                slave.status = "failed to reset replication"
                slave.severity = Severity.WARNING.value
                slave.ind_started = True
                slave.save()
                continue

            slave.status = ""
            slave.save()

        slaves = c123_node.objects.filter(cluster = app_primary.cluster, type = app_primary.type, cluster_role = ClusterRoles.STANDBY.value, severity = Severity.OK.value, ind_started = True,  ind_deleted = False)
        if not slaves.exists():
            print(error_label, f'could not find replacement candidates for primary {app_primary.name}')
            return

        master_host = None
        master = None
        master_file = None
        master_position = None
        master_client = None
        for slave in slaves:
            slave_client = create_mysql_connection(slave)
            if isinstance(slave_client, tuple):
                continue

            try:
                cur = slave_client.cursor(dictionary=True)
                cur.execute(f"SHOW {'BINARY LOG' if float(mysql_version) > 8.0 and mysql_app in type else 'MASTER'} STATUS;")
                res = cur.fetchone()
                if res:
                    master_file = res.fget('File')
                    master_position = res.get('Position')
                else:
                    slave.status = "failed to obtain binlog info"
                    slave.severity = Severity.WARNING.value
                    slave.ind_started = True
                    slave.save()
                    print(error_label, slave.name, slave.status)
                    continue
            except Exception as e:
                error = e.strerror if hasattr(e, 'strerror') else str(e)
                print(error_label, slave.name,error)
                slave.status = "failed to obtain binlog info"
                slave.severity = Severity.WARNING.value
                slave.ind_started = True
                slave.save()
                continue

            if master_file and master_position:
                master_host = slave.host
                master = slave
                master_client = slave_client
                break
        if not master_client:
            print(error_label, f'could not find replacement candidates for primary (no slave) {app_primary.name}')
            return

        master.cluster_role = ClusterRoles.PRIMARY.value
        master.ex_params[link_field] = master.pk
        master.save()
        slaves = c123_node.objects.filter(cluster = app_primary.cluster, cluster_role = ClusterRoles.STANDBY.value, type = app_primary.type, ind_deleted = False)
        if slaves.exists():
            master.severity = Severity.WARNING.value
            master.save()
            for slave in slaves:
                master.status = f'synchronizing with replica {slave.name}'
                master.save()
                slave_host = slave.host
                #if slave was slave before - using its credentials
                if slave.ex_params.get('rep_user'):
                    rep_user = slave.ex_params.get('rep_user')
                    rep_pwd = slave.ex_params.get('rep_pwd')
                #if its former initial primary - inherit credentials of promoted standby
                else:
                    rep_user = master.ex_params.get('rep_user')
                    rep_pwd = master.ex_params.get('rep_pwd')

                slave_setup_error = False
                try:
                    cur = master_client.cursor()
                    if type in mysql_app:
                        password_wrapper = f"PASSWORD('{rep_pwd}')"
                    else:
                        password_wrapper = f"'{rep_pwd}'"
                    commands = [
                        f"CREATE USER IF NOT EXISTS '{rep_user}'@'%.%.%.%' IDENTIFIED {authentication_plugin} {password_wrapper} {'REQUIRE SSL' if use_ssl else ''}",
                        f"ALTER USER '{rep_user}'@'%.%.%.%' IDENTIFIED {authentication_plugin} {password_wrapper} {'REQUIRE SSL' if use_ssl else ''}",
                        f"GRANT ALL PRIVILEGES ON *.* TO '{rep_user}'@'%.%.%.%' WITH GRANT OPTION",
                        f"GRANT REPLICATION SLAVE ON *.* TO '{rep_user}'@'%.%.%.%'",
                        'FLUSH PRIVILEGES'
                    ]
                    for command in commands:
                        cur.execute(command)
                except Exception as e:
                    error = e.strerror if hasattr(e, 'strerror') else str(e)
                    print(error_label, master.name,error)
                    master.severity = Severity.WARNING.value
                    master.ind_started = True
                    master.status = f'failed to create replication user of replica {slave.name}'
                    slave_setup_error = True
                    break

                if slave_setup_error:
                    continue

                slave_client = create_mysql_connection(slave)
                if isinstance(slave_client, tuple):
                    continue

                commands = [
                    f"""CHANGE {master_type} {master_label} TO {master_label}_HOST=\'{master_host.name}\', {master_label}_USER=\'{rep_user}\', {master_label}_PORT={master.port}, {f"{master_label}_SSL=1, {master_label}_SSL_VERIFY_SERVER_CERT=0, {master_label}_SSL_CAPATH='{master.dir_123Cluster}', {master}_SSL_CA='{master.dir_123Cluster}/ca.pem', {master}_SSL_CERT='{master.dir_123Cluster}/client-cert.pem', {master}_SSL_KEY='{master.dir_123Cluster}/client-key.pem'," if use_ssl else ""} {master_label}_PASSWORD=\'{rep_pwd}\', {master_label}_LOG_FILE=\'{master_file}\', {master_label}_LOG_POS={master_position};"""
                    f"START {slave_label};"
                ]
                try:
                    cur = slave_client.cursor()
                    for command in commands:
                        cur.execute(command)
                except Exception as e:
                    error = e.strerror if hasattr(e, 'strerror') else str(e)
                    print(error_label, slave.name,error)
                    slave.severity = Severity.WARNING.value
                    slave.status = f'failed to link to master {master.name}'
                    slave.ind_started = True
                    slave.save()
                    slave_setup_error = True
                    break

                ssh_client = create_ssh_connection(slave_host.name, slave_host.username, slave_host.password, slave_host.private_key, slave_host.ssh_port)
                if isinstance(ssh_client, tuple):
                    slave.severity = Severity.NOT_OK.value
                    slave.status = Status.OFFLINE.value
                else:
                    slave.severity = Severity.OK.value
                    slave.status = Status.OK.value
                slave.save()


                commands = [
                    f"sudo grep -qxF '[mysqld]' '{conf_path}' && grep -qxF '{conf_variable}=ON' '{conf_path}' || sudo sed -i '/\[mysqld\]/a {conf_variable}=ON' '{conf_path}'",
                    f'sudo systemctl restart {type}{appendix}',
                ]
                for command in commands:
                    jsn = json.loads(remote_command(ssh_client, command, True), strict = False)
                    if jsn['stderr'] != '' and 'error' in jsn['stderr'].lower():
                        print(error_label, slave.name, jsn['stderr'])
                        slave_setup_error = True
                        break

                if slave_setup_error:
                    slave.status = 'failed to made read-only'
                    slave.severity = Severity.WARNING.value
                    slave.save()
                    continue

                slave.severity = Severity.OK.value
                slave.status = Status.OK.value
                slave.ind_started = True
                slave.save()

            master.severity = Severity.OK.value
            master.status = Status.OK.value
            master.ind_started = True
            master.save()

    slaves = c123_node.objects.filter(cluster = app_primary.cluster, cluster_role = ClusterRoles.STANDBY.value, type = app_primary.type,  ind_deleted = False).exclude(status = Status.DROPPING.value)
    for slave in slaves:
        slave.ex_params['lag'] = get_mysql_fork_standby_lag(slave)
        slave.save()

@close_old_connections
def drop_mariadb_node(node, final_backup = False):
    cluster_ids = c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).values_list('id', flat=True)
    existing_drop = Task.objects.filter(target_id__in = cluster_ids, task = 'drop_mariadb_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if existing_drop:
        raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    node.status = Status.DROPPING.value
    node.save()
    if node.cluster_role != ClusterRoles.PRIMARY.value:
        final_backup = False
    if node.cluster_role != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MARIADB.value)
        keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
        # create a unique string with the date
        now = timezone.now().strftime('%d%m%Y%H%M%S')
        config = SystemConfig.objects.filter(db_type = DBType.MARIADB.value).first()
        backup_path = config.backups_path
        ex_params = {
            'size': 0,
            'duration': 0,
            'belong_to': DBType.MARIADB.value,
            'node_version': node.version,
            'compress': True,
            'root_pwd': node.ex_params.get('mariadb_root_password'),
            'modified_date': '',
            'backup_path': f'{backup_path}/{node.host.name}_{node.port}/{now}'
        }

        final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
        node = node,
        type = Inventory.BACKUP.value,
        created_by = node.created_by,
        is_sched = False,
        keep_days = keep_days,
        ex_params = ex_params
        )
    cleanup_logs(node.pk, node.type)
    update_host_usage(node.host)
    from django.conf import settings
    # check if node the only one in cluster:
    other_nodes = c123_node.objects.filter(cluster = node.cluster, type = DBType.MARIADB.value, ind_deleted = False).exclude(pk = node.pk)

    has_other_nodes = other_nodes.exists()
    if not has_other_nodes:
        job = settings.scheduler.get_job(job_id=f"update_mysql_fork_roles_{node.cluster.id}")
        if job:
            job.remove()

    node_jobs = [i for i in settings.scheduler.get_jobs() if node.name in i.id]
    for job in node_jobs:
        job.remove()

    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = node.created_by
    standbys = [{"node": object_to_json(i), "host": object_to_json(i.host)} for i in other_nodes.filter(cluster_role = ClusterRoles.STANDBY.value)]
    tmp_data = {'node_data': node_data, 'host_data': host_data, 'standbys': standbys, 'created_by': created_by.id}
    if node.cluster_role == ClusterRoles.STANDBY.value:
        primary =  c123_node.objects.filter(ex_params__mariadb_master_id=node.ex_params.get('mariadb_master_id'), type = DBType.MARIADB.value, cluster_role=ClusterRoles.PRIMARY.value, ind_deleted = False).first()
        if primary:
            tmp_data['primary_data'] = object_to_json(primary)
            tmp_data['host_primary'] = c123_HostSerializer(primary.host).data

    tmp_data['final_backup'] = c123_inventory_serializer(final_backup).data if final_backup else final_backup
    tmp_data['backup_host'] = c123_HostSerializer(app_host).data
    tmp_data = json.dumps(tmp_data)
    summary = generate_license_summary()
    from clusterCore.api.licensing.views import LicenseType
    revoke_license(node)
    spread_licenses_by_type(LicenseType.MARIADB.value, summary.mariadb)
    task = Task.objects.create(task='drop_mariadb_node_task',
                                displayment='drop {} {}'.format(node.cluster_role.lower(), node.name),
                                target_type=DBType.MARIADB.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    return task



@close_old_connections
def update_elasticsearch_roles(app_primary):

    if app_primary.status in [Status.STOPPING.value, Status.STOPPED.value]:
        return

    should_skip = Task.objects.filter(task__in = ['switchover_elasticsearch_node_task', 'drop_elasticsearch_node_task'], target_id = app_primary.pk, status__in = [TaskState.PENDING.value, TaskState.RUNNING.value]).exists()
    if should_skip:
        return

    standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.ELASTIC.value, cluster = app_primary.cluster, ind_deleted = False)
    if standbys.count() > 0:
        is_primary = check_if_elastic_primary(app_primary)
        #primary is standby - strict equals, because can return string error
        if is_primary is False:
            #searching for new primary in primary standbys
            for standby in standbys:
                is_primary = check_if_elastic_primary(app_primary)
                #standby is primary
                if is_primary is True:
                    standby.cluster_role = ClusterRoles.PRIMARY.value
                    standby.ex_params["elasticsearch_master_id"] = standby.id
                    standby.save()
                    others = c123_node.objects.filter(type = DBType.ELASTIC.value, cluster = app_primary.cluster, ind_deleted = False).exclude(cluster_role = ClusterRoles.PRIMARY.value)
                    for other in others:
                        other.ex_params["elasticsearch_master_id"] = standby.id
                        other.save()

                    #downgrade former primary to standby of new primary
                    app_primary.ex_params["elasticsearch_master_id"] = standby.id
                    app_primary.cluster_role = ClusterRoles.STANDBY.value
                    app_primary.save()
                    return

@close_old_connections
def remove_pg_slots(cluster):
    standbys = c123_node.objects.filter(type = DBType.POSTGRES.value, cluster = cluster, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False).exclude(status = Status.DROPPING.value)
    for standby in standbys:
        conn = get_postgres_connection(standby)
        if not isinstance(conn, tuple):
            with conn:
                with conn.cursor() as cur:
                    slot_name = "slot_123cluster"
                    cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name = %s", (slot_name,))
                    slot_exists = cur.fetchone()
                    if slot_exists:
                        cur.execute("SELECT pg_drop_replication_slot(%s)", (slot_name,))
                        conn.commit()
        else:
             logging.error(f"remove_pg_slots_{cluster.id}: {conn[1]}")

@close_old_connections
def update_pg_roles(app_primary):
    try:
        if app_primary.status in [Status.STOPPING.value, Status.STOPPED.value]:
            return
        cluster = app_primary.cluster
        # cluster initially always have primary
        if not cluster.ex_params:
            cluster.ex_params = {'primary_count': 1}
            cluster.save()

        should_stop = Task.objects.filter(task__in=['install_postgres_standalone_task', 'drop_pg_node_task', 'pg_cluster_primary_search_task', 'pg_split_brain_recovery_task', 'pg_node_restore_task', 'switchover_repmgr_task', 'pg_upgrade_node_task'], status__in = [TaskState.PENDING.value, TaskState.RUNNING.value], target_id = app_primary.pk).exists()
        if should_stop:
            return
        # now we want reference to all "primaries" of cluster, for the split brain task
        primary_ids = []
        # actually no need to check if standalone,
        # if we have more than 1 of standalones/primaries in cluster - thats split brain anyway
        cluster_nodes = c123_node.objects.filter(type = DBType.POSTGRES.value, cluster = cluster, ind_deleted = False).exclude(status = Status.DROPPING.value)
        cluster_nodes_id = cluster_nodes.values_list('pk', flat=True)
        # no cluster role here, just "others than current target"
        other_than_target = cluster_nodes.exclude(pk = app_primary.pk)
        other_nodes_exists = other_than_target.exists()
        is_primary = check_if_pg_primary(app_primary, in_cluster = other_nodes_exists)
        if is_primary:
            primary_count = 1
            primary_ids.append(app_primary.pk)
        else:
            primary_count = 0

        cluster.ex_params["primary_count"] = primary_count
        cluster.save()

        # then exit after checking first node
        if not other_nodes_exists:
            return

        new_primary = None
        for node in other_than_target:
            is_primary = check_if_pg_primary(node, in_cluster=True)
            #likely primary here, but no 100% guarantee
            if is_primary:
                new_primary = node
                primary_count += 1
                cluster.ex_params["primary_count"] = primary_count
                cluster.save()
                # if split brain
                if primary_count > 1:
                    # assume we could use that status more than once
                    node.status = Status.SPLIT_BRAIN.value
                    node.severity = Severity.NOT_OK.value
                    node.save()
                    primary_ids.append(node.pk)
                    created_by = app_primary.created_by
                    cluster_data = [{"host": object_to_json(n.host), "node": object_to_json(n)} for n in cluster_nodes]
                    tmp_data = json.dumps({'primary_ids': primary_ids,  'cluster_data': cluster_data, 'created_by': created_by.id})
                    Task.objects.filter(task='pg_split_brain_recovery_task', status = TaskState.ON_DEMAND.value, target_id__in=cluster_nodes_id).delete()
                    Task.objects.create(task='pg_split_brain_recovery_task',
                                            displayment=f'fixing split brain for cluster of {app_primary.name}',
                                            target_type=DBType.POSTGRES.value,
                                            target_id=app_primary.pk,
                                            org_id=app_primary.org_id,
                                            schedule=Schedule.NOW.value,
                                            status=TaskState.PENDING.value,
                                            parameters=tmp_data,
                                            created_by=created_by)

        if new_primary and primary_count == 1:
            new_primary.cluster_role = ClusterRoles.PRIMARY.value
            new_primary.ex_params["pg_primary_id"] = node.pk
            new_primary.ex_params["lag"] = "N/A (node is primary)"
            new_primary.save()
            others = cluster_nodes.exclude(pk = new_primary.pk)
            for other in others:
                other.ex_params["pg_primary_id"] = new_primary.pk
                other.cluster_role = ClusterRoles.STANDBY.value
                other.save()
            cluster.ex_params["new_basebackup"] = True
            cluster.save()

            conn = get_postgres_connection(new_primary)
            if not isinstance(conn, tuple):
                with conn:
                    with conn.cursor() as cur:
                        slot_name = "slot_123cluster"
                        cur.execute("SELECT slot_name FROM pg_replication_slots WHERE slot_name = %s", (slot_name,))
                        slot_exists = cur.fetchone()
                        if not slot_exists:
                            cur.execute("SELECT pg_create_physical_replication_slot(%s)", (slot_name,))
                            conn.commit()
            else:
                logging.error(f"update_pg_roles_{cluster.id}: {conn[1]}")

        # if primary are not found
        if primary_count == 0:
            created_by = app_primary.created_by
            cluster_data = [{"host": object_to_json(n.host), "node": object_to_json(n)} for n in cluster_nodes]
            tmp_data = json.dumps({'old_primary_id': app_primary.pk,  'cluster_data': cluster_data, 'created_by': created_by.id})
            Task.objects.filter(task='pg_cluster_primary_search_task', status = TaskState.ON_DEMAND.value, target_id__in=cluster_nodes_id).delete()
            Task.objects.create(task='pg_cluster_primary_search_task',
                                    displayment=f'searching of new primary for cluster of {app_primary.name}',
                                    target_type=DBType.POSTGRES.value,
                                    target_id=app_primary.pk,
                                    org_id=app_primary.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=created_by)

        #lag take on changed set (assume that cluster roles here correctly set already)
        standbys = c123_node.objects.filter(cluster_role = ClusterRoles.STANDBY.value, type = DBType.POSTGRES.value, cluster = app_primary.cluster, ind_deleted = False).exclude(status = Status.DROPPING.value)
        for standby in standbys:
            standby.ex_params["lag"] = get_pg_standby_lag(standby)
            standby.save()

    except Exception as e:
        logging.error(f"update_pg_roles_{cluster.id}: {e}")

def create_api_command(data, access_token, endpoint):
    headers = [
        f'-H "Authorization: {access_token}"',
        '-H "Content-Type: application/json"',
        '-H "Accept: application/json"',
    ]
    escaped_data = json.dumps(data, separators=(',', ':')).replace('"', '\\"').replace("'", "\\'")
    api_command = f'curl -v {" ".join(headers)} -d "{escaped_data}" {endpoint}'
    return api_command

def append_status_cmd(request, task, custom_notification = None):
    relative_url = reverse('check_task_status')
    check_task_status_endpoint = request.build_absolute_uri(relative_url)
    auth_header = request.META.get('HTTP_AUTHORIZATION')
    status_cmd = create_api_command({"task_id": task.task_id}, auth_header, check_task_status_endpoint)
    notification = custom_notification if custom_notification else task.displayment
    response = JsonResponse({"notification": f"{notification} started", "task": task.displayment}, status = 200)
    response['X-123Cluster-Task-Status'] = status_cmd
    return response

def format_password(password):
    from django.conf import settings
    import base64
    encrypt = getattr(settings, 'ENCODE_PASSWORDS', None)
    if encrypt:
        return base64.b64decode(password).decode('utf-8')
    return password



@close_old_connections
def drop_pg_node(node, final_backup = False):
    cluster_ids = c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).values_list('id', flat=True)
    existing_drop = Task.objects.filter(target_id__in = cluster_ids, task = 'drop_pg_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if existing_drop:
        raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    node.status = Status.DROPPING.value
    node.save()
    if node.cluster_role != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
        keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
        # create a unique string with the date
        now = timezone.now().strftime('%d%m%Y%H%M%S')
        config = SystemConfig.objects.filter(db_type = DBType.POSTGRES.value).first()
        backup_path = config.backups_path
        ex_params = {
            'size': 0,
            'duration': 0,
            'belong_to': DBType.POSTGRES.value,
            'node_version': node.version,
            'compress': True,
            'backup_type': 'basebackup',
            'modified_date': '',
            'backup_path': f'{backup_path}/basebackups/{node.host.name}_{node.port}/{now}'
        }

        final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
        node = node,
        type = Inventory.BACKUP.value,
        created_by = node.created_by,
        is_sched = False,
        keep_days = keep_days,
        ex_params = ex_params
        )
    cleanup_logs(node.pk, node.type)
    update_host_usage(node.host)

    from django.conf import settings
    # check if node the only one in cluster
    other_nodes = c123_node.objects.filter(type = DBType.POSTGRES.value,  cluster = node.cluster, ind_deleted = False).exclude(pk = node.pk)
    has_other_nodes = other_nodes.exists()
    if not has_other_nodes:
        job = settings.scheduler.get_job(job_id=f"update_pg_roles_{node.cluster.id}")
        if job:
            job.remove()
        job = settings.scheduler.get_job(job_id=f"remove_pg_slots_{node.cluster.id}")
        if job:
            job.remove()

    node_jobs = [i for i in settings.scheduler.get_jobs() if node.name in i.id]
    for job in node_jobs:
        job.remove()

    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = node.created_by
    standby_data = [{'host': object_to_json(i.host), 'node': object_to_json(i)} for i in other_nodes.filter(cluster_role = ClusterRoles.STANDBY.value)]
    tmp_data = {'node_data': node_data, 'host_data': host_data, 'standby_data': standby_data, 'final_backup': c123_inventory_serializer(final_backup).data if final_backup else final_backup, 'backup_host': c123_HostSerializer(app_host).data, 'created_by': created_by.id}
    haproxies = c123_node.objects.filter(cluster = node.cluster, type = DBType.HAPROXY.value, ind_deleted = False)
    haproxy_exists = haproxies.exists()
    tmp_data['haproxy_exists'] = haproxy_exists
    if haproxy_exists:
        nodes = c123_node.objects.filter(type = DBType.POSTGRES.value, ind_deleted = False, cluster = node.cluster).exclude(pk = node.id).exclude(type = DBType.HAPROXY.value)
        node_list = [{'node_id': i.pk,  "host_ip": i.host.name, "node_port": i.port} for i in nodes]
        tmp_data['node_list'] = node_list
        tmp_data['haproxy_hosts'] = []
        known_hosts = []
        for haproxy in haproxies:
            if haproxy.host not in known_hosts:
                tmp_data['haproxy_hosts'].append(object_to_json(haproxy.host))
                known_hosts.append(haproxy.host)

    summary = generate_license_summary()
    from clusterCore.api.licensing.views import LicenseType
    revoke_license(node)
    spread_licenses_by_type(LicenseType.POSTGRES.value, summary.postgresql)
    task = Task.objects.create(task='drop_pg_node_task',
                                displayment='drop {} {}'.format(node.cluster_role.lower(), node.name),
                                target_type=DBType.POSTGRES.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=json.dumps(tmp_data),
                                created_by=created_by)
    return task

@close_old_connections
def drop_pdb(cdb, pdb, options = "justDrop"):
    existing_drop = Task.objects.filter(target_id = pdb.id, task = 'drop_pdb_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if existing_drop:
        raise APIException409(f"Already runs drop for pdb {pdb.name}")
    pdb.status = Status.DROPPING.value
    pdb.save()
    cleanup_logs(pdb.pk, pdb.type)
    update_host_usage(pdb.host)

    from django.conf import settings
    if Status.DETACHED.value in pdb.status:
        pdb.ind_deleted = True
        pdb.save()
        return  f"pdb {pdb.name} is already deleted"

    node_jobs = [i for i in settings.scheduler.get_jobs() if pdb.name in i.id]
    for job in node_jobs:
        job.remove()

    cdb_data = object_to_json(cdb)
    pdb_data = object_to_json(pdb)
    host_data = c123_HostSerializer(pdb.host).data
    created_by = pdb.created_by
    tmp_data = json.dumps({'cdb_data': cdb_data, 'pdb_data': pdb_data, 'host_data': host_data,  'option': options, 'created_by': created_by.id})
    summary = generate_license_summary()
    from clusterCore.api.licensing.views import LicenseType
    revoke_license(pdb)
    spread_licenses_by_type(LicenseType.ORACLE.value, summary.oracle)
    task = Task.objects.create(task='drop_pdb_task',
                                displayment=f'drop {pdb.name}',
                                target_type=DBType.ORACLE_PDB.value,
                                target_id=pdb.id,
                                org_id=pdb.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    return task

@close_old_connections
def drop_elasticsearch_node(node, final_backup = False):
    cluster_ids = c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).values_list('id', flat=True)
    existing_drop = Task.objects.filter(target_id__in = cluster_ids, task = 'drop_elasticsearch_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if existing_drop:
        raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    node.status = Status.DROPPING.value
    node.save()
    if node.cluster_role != ClusterRoles.PRIMARY.value:
        final_backup = False
    if final_backup:
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.ELASTIC.value)
        keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
        # create a unique string with the date
        now = timezone.now().strftime('%d%m%Y%H%M%S')
        config = SystemConfig.objects.filter(db_type = DBType.ELASTIC.value).first()
        backup_path = config.backups_path
        ex_params = {
            'size': 0,
            'duration': 0,
            'belong_to': DBType.ELASTIC.value,
            'node_version': node.version,
            'compress': True,
            'backup_type':  'cluster_snapshot',
            'modified_date': '',
            'backup_path': f'{backup_path}/{node.host.name}_{node.port}/{now}'
        }

        final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
        node = node,
        type = Inventory.BACKUP.value,
        created_by = node.created_by,
        is_sched = False,
        keep_days = keep_days,
        ex_params = ex_params
        )
    cleanup_logs(node.pk, node.type)
    update_host_usage(node.host)

    from django.conf import settings
    # check if node the only one in cluster
    cluster_nodes = c123_node.objects.filter(cluster = node.cluster,type = DBType.ELASTIC.value, ind_deleted  = False)
    has_other_nodes = cluster_nodes.exclude(pk = node.pk).exists()
    if not has_other_nodes:
            job = settings.scheduler.get_job(job_id=f"update_elasticsearch_roles_{node.cluster.id}")
            if job:
                job.remove()

    node_jobs = [i for i in settings.scheduler.get_jobs() if node.name in i.id]
    for job in node_jobs:
        job.remove()

    created_by = node.created_by
    tmp_data = json.dumps({'node_id': node.pk, 'cluster_data': [{'host': object_to_json(n.host), 'node': object_to_json(n)} for n in cluster_nodes], 'final_backup': c123_inventory_serializer(final_backup).data if final_backup else final_backup, 'backup_host': c123_HostSerializer(app_host).data, 'created_by': created_by.id})
    summary = generate_license_summary()

    from clusterCore.api.licensing.views import LicenseType
    revoke_license(node)
    spread_licenses_by_type(LicenseType.ELASTIC.value, summary.elasticsearch)
    task = Task.objects.create(task='drop_elasticsearch_node_task',
                                displayment='drop {} {}'.format(node.cluster_role.lower(), node.name),
                                target_type=DBType.ELASTIC.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    return task

@close_old_connections
def drop_mongo_node(node, final_backup = False):
    cluster_ids = c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).values_list('id', flat=True)
    existing_drop = Task.objects.filter(target_id__in = cluster_ids, task__in = ['delete_mongo_standalone_node_task', 'delete_mongo_replicaset_node_task'], status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
    if existing_drop:
        raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")
    node.status = Status.DROPPING.value
    node.save()
    if final_backup and node.cluster_role != ClusterRoles.ARBITER.value:
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MONGO.value)
        keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
        # create a unique string with the date
        now = timezone.now().strftime('%d%m%Y%H%M%S')
        config = SystemConfig.objects.filter(db_type = DBType.MONGO.value).first()
        backup_path = config.backups_path
        ex_params = {
            'size': 0,
            'duration': 0,
            'belong_to': DBType.MONGO.value,
            'node_version': node.version,
            'compress': True,
            'modified_date': '',
            'backup_path': f'{backup_path}/{node.host.name}_{node.port}/{now}'
        }

        final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
        node = node,
        type = Inventory.BACKUP.value,
        created_by = node.created_by,
        is_sched = False,
        keep_days = keep_days,
        ex_params = ex_params
        )
    else:
        final_backup = False
    cleanup_logs(node.pk, node.type)
    update_host_usage(node.host)

    from django.conf import settings
    other_secondaries = c123_node.objects.filter(cluster = node.cluster, type=DBType.MONGO.value, cluster_role=ClusterRoles.STANDBY.value, ind_deleted = False).exclude(pk = node.pk)
    has_other_secondaries = other_secondaries.exists()
    arbiters = c123_node.objects.filter(cluster=node.cluster, type=DBType.MONGO.value, cluster_role=ClusterRoles.ARBITER.value, ind_deleted = False)
    arbiters_id = arbiters.values_list('id', flat=True)
    if not c123_node.objects.filter(cluster = node.cluster, type=DBType.MONGO.value, ex_params__replicaset_name = node.ex_params.get('replicaset_name'), ind_deleted = False).exclude(cluster_role = ClusterRoles.ARBITER.value).exclude(pk = node.pk):
        job = settings.scheduler.get_job(job_id=f"update_mongo_roles_{node.cluster.id}")
        if job:
            job.remove()

    node_jobs = [i for i in settings.scheduler.get_jobs() if node.name in i.id]
    for job in node_jobs:
        job.remove()

    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = node.created_by
    tmp_data = {'node_data': node_data, 'host_data': host_data, 'created_by': created_by.id}
    secondaries_data = []
    standbys = []
    for secondary in other_secondaries:
        secondaries_data.append({"id":secondary.id, "name": secondary.name})
        standbys.append({"host": object_to_json(secondary.host), "node": object_to_json(secondary)})

    tmp_data['secondaries_data'] = secondaries_data
    tmp_data['standbys']  = standbys
    arbiter_data = []
    for arbiter in arbiters:
        arbiter_data.append({"host": object_to_json(arbiter.host), "node": object_to_json(arbiter)})
    tmp_data['arbiters_data'] = arbiter_data
    tmp_data['arbiters_id'] = list(arbiters_id)
    if ClusterRoles.BALANCER.value not in node.cluster_role:
        primary = c123_node.objects.filter(cluster = node.cluster, cluster_role = ClusterRoles.PRIMARY.value, type=DBType.MONGO.value, ind_deleted = False).first()
        if not primary:
            raise APIException410(f"primary of node {node.name} not available")

        tmp_data['primary_data'] = object_to_json(primary)
        tmp_data['host_primary'] = object_to_json(primary.host)

    tmp_data['final_backup'] =  c123_inventory_serializer(final_backup).data if final_backup else final_backup
    tmp_data['backup_host'] = c123_HostSerializer(app_host).data
    tmp_data = json.dumps(tmp_data)
    summary = generate_license_summary()
    node.ind_license = False
    node.save()

    from clusterCore.api.licensing.views import LicenseType
    revoke_license(node)
    spread_licenses_by_type(LicenseType.MONGO.value, summary.mongo)
    task_name = 'delete_mongo_replicaset_node_task'

    task = Task.objects.create(task=task_name,
                                displayment=f'drop mongo {node.cluster_role.lower()} {node.name}',
                                target_type=DBType.MONGO.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    return task


def create_mysql_connection(node, database = 'mysql'):
    try:

        result = check_host_connection(node)
        if node.type == DBType.MARIADB.value:
            ssl = node.ex_params.get('ssl')
            client = mariadb.connect(host=node.host.name, port=int(node.port), user=node.database_user, password = node.database_pwd, database = database, connect_timeout = Timeout.BACKEND_MARIADB.value, ssl = ssl)
        else:
            client = mysql.connector.connect(host=node.host.name, port=int(node.port), user=node.database_user, password = node.database_pwd, database = database, connection_timeout = Timeout.BACKEND_MYSQL.value)
        node.status = Status.OK.value
        node.severity = Severity.OK.value
        node.ind_started = True
        node.save()
        return client
    except Exception as e:
        if not isinstance(result, tuple):
            if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                node.status = Status.STOPPED.value
            else:
                node.status = Status.DOWN.value

            node.severity = Severity.NOT_OK.value
            node.ind_started = False
            node.save()
            c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        logging.error(f"create_mysql_connection: {e}")
        return (False, e)



def create_ssh_connection(hostname, user, pwd, pkey, prt=22):
    """ssh connection creation helper function"""

    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    try:
        if pkey and isinstance(pkey, str) and not pkey.isspace():
            if '\n' not in pkey:
               pkey = format_rsa_key(pkey)
            private_key = paramiko.RSAKey.from_private_key(StringIO(pkey))
            client.connect(hostname=hostname, port=prt, username=user, pkey=private_key, look_for_keys=False, auth_timeout=15, banner_timeout=15, timeout=15)
        elif pwd and isinstance(pwd, str) and not pwd.isspace():
            pwd = pwd.strip()
            client.connect(hostname=hostname, port=prt, username=user, password=pwd, banner_timeout=15, timeout=15)
        else:
            return (False, f'cannot connect to {hostname}')

        return client

    except (paramiko.ssh_exception.NoValidConnectionsError, EOFError) as e:
        logging.error(f"create_ssh_connection: {e}")
        return (False, e)
    except paramiko.ssh_exception.SSHException as e:
        logging.error(f"create_ssh_connection: {e}")
        return (False, e)
    except Exception as e:
        logging.error(f"create_ssh_connection: {e}")
        return (False, e)

@close_old_connections
def drop_redis_node(node, data_only = False, final_backup = False):
        cluster_ids = c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).values_list('id', flat=True)
        existing_drop = Task.objects.filter(target_id__in = cluster_ids, task = 'drop_redis_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
        if existing_drop:
            raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        node.status = Status.DROPPING.value
        node.save()
        if final_backup and node.cluster_role != ClusterRoles.SENTINEL.value:
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.REDIS.value)
            keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
            # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            config = SystemConfig.objects.filter(db_type = DBType.REDIS.value).first()
            backup_path = config.backups_path
            ex_params = {
                'size': 0,
                'duration': 0,
                'belong_to': DBType.REDIS.value,
                'backup_type': 'rdb',
                'node_version': node.version,
                'compress': True,
                'modified_date': '',
                'backup_path': f'{backup_path}/{node.host.name}_{node.port}/{now}'
            }

            final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
            node = node,
            type = Inventory.BACKUP.value,
            created_by = node.created_by,
            is_sched = False,
            keep_days = keep_days,
            ex_params = ex_params
            )
        update_host_usage(node.host)
        has_slave = c123_node.objects.filter(type = DBType.REDIS.value, cluster_role = ClusterRoles.STANDBY.value, ex_params__slave_of=node.ex_params.get("slave_of"), ind_deleted = False).exclude(pk = node.pk).exists()

        cleanup_logs(node.pk, node.type)
        node_data = object_to_json(node)
        host_data = c123_HostSerializer(node.host).data
        created_by = node.created_by

        from django.conf import settings
        node_jobs = [i for i in settings.scheduler.get_jobs() if node.name in i.id]
        for job in node_jobs:
            job.remove()

        other_nodes = [{"host": object_to_json(i.host), "node": object_to_json(i)} for i in c123_node.objects.filter(type = DBType.REDIS.value, ind_deleted = False, cluster = node.cluster).exclude(pk = node.pk).exclude(cluster_role = ClusterRoles.SENTINEL.value)]
        tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'has_slave': has_slave, 'other_nodes': other_nodes, 'created_by': created_by.id, 'data_only': data_only, 'final_backup': c123_inventory_serializer(final_backup).data if final_backup else final_backup, 'backup_host': c123_HostSerializer(app_host).data})
        summary = generate_license_summary()

        from clusterCore.api.licensing.views import LicenseType
        revoke_license(node)
        spread_licenses_by_type(LicenseType.REDIS.value, summary.redis)
        task = Task.objects.create(task='drop_redis_node_task',
                                    displayment='drop {} {}'.format(node.cluster_role.lower(), node.name),
                                    target_type=DBType.REDIS.value,
                                    target_id=node.id,
                                    org_id=node.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                   created_by=created_by)
        return task

@close_old_connections
def drop_mssql_node(node):
    node.status = Status.DROPPING.value
    node.save()
    cleanup_logs(node.pk, node.type)
    update_host_usage(node.host)
    other_standbys = c123_node.objects.filter(type = DBType.MSSQL.value, cluster_role = ClusterRoles.STANDBY.value, ex_params__mssql_master_id=node.ex_params.get("mssql_master_id"), ind_deleted = False).exclude(pk = node.pk)

    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = node.created_by
    standbys = [{"node": object_to_json(i), "host": object_to_json(i.host)} for i in other_standbys]
    tmp_data = json.dumps({'node_data': node_data, 'host_data': host_data, 'standbys': standbys, 'created_by': created_by.id})
    summary = generate_license_summary()

    from clusterCore.api.licensing.views import LicenseType
    revoke_license(node)
    spread_licenses_by_type(LicenseType.MSSQL.value, summary.mssql)
    task = Task.objects.create(task='unregister_sqlserver_instance_task',
                                displayment='unregister {} {}'.format(node.cluster_role.lower(), node.name),
                                target_type=DBType.MSSQL.value,
                                target_id=node.id,
                                org_id=node.org_id,
                                schedule=Schedule.NOW.value,
                                status=TaskState.PENDING.value,
                                parameters=tmp_data,
                                created_by=created_by)
    return task

def remote_command_win(session, cmd, wait_for_output,  shell = 'cmd') -> dict:
    json_str = {'stdout': '', 'stderr': '', 'excode': 0}
    try:
        if shell == 'ps':
            res = session.run_ps(cmd)
        else:
            res = session.run_cmd(cmd)

        print('remote_command_win {}'.format(cmd))
        if wait_for_output:
            json_str['stdout'] =  res.std_out.decode().strip()
            json_str['stderr'] =  res.std_err.decode().strip()
            json_str['excode'] =  res.status_code
    except Exception as e:
        print("remote_command_win:",  str(e))
        json_str['stdout'] = ''
        json_str['stderr'] = e.strerror if hasattr(e, 'strerror') else str(e)
        json_str['excode'] = 1

    return json.dumps(json_str)

def connect_to_oracle_database(**kwargs):
    """return a oracledb connection object
    kwargs contain the CDB object from which the
    host, username and password are used to create
    the connection"""


    try:
        cdb = kwargs.pop('cdb')
        result = check_host_connection(cdb)
        dsn_tns = oracledb.makedsn(cdb.host.name, cdb.port, cdb.name)
        conn = oracledb.connect(user=cdb.database_user, password=cdb.database_pwd, tcp_connect_timeout=Timeout.BACKEND_ORACLE.value, dsn=dsn_tns, mode=oracledb.SYSDBA)
        cdb.severity = Severity.OK.value
        cdb.status = Status.OK.value
        cdb.ind_started = True
        cdb.save()
        return conn
    except Exception as e:
        if not isinstance(result, tuple):
            if cdb.status in [Status.STOPPING.value, Status.STOPPED.value]:
                cdb.status = Status.STOPPED.value
            else:
                cdb.status = Status.DOWN.value

            cdb.severity = Severity.NOT_OK.value
            cdb.ind_started = False
            cdb.save()

            logging.error(f"connect_to_oracle_database: {e}")
            if "ORA-12545" in str(e):
                cdb.ind_deleted = True
                cdb.save()
                return {"message": str(e), "verbose": True}

        raise APIException401(str(e))


def create_winrm_connection(hostname, user, pwd, prt=5985, protocol = 'http', transport = 'ntlm', endpoint = 'wsman',  server_cert_validation='ignore'):
    """windows remote connection creation helper function"""

    try:
        return Session(target = f'{protocol}://{hostname}:{prt}/{endpoint}', auth=(user, pwd), **{'transport': transport, 'server_cert_validation': server_cert_validation})
    except Exception as e:
        logging.error(f"create_winrm_connection: {e}")
        return (False, e)


def local_command(command, wait=True):
    """
    Runs a shell command on the local machine and returns JSON-encoded dictionary.
    
    :param command: The shell command to execute (as a string or list).
    :param wait: Whether to wait for command completion.
    :return: JSON string with stdout, stderr, and excode.
    """
    try:
        env = os.environ.copy()
        # Ensure essential system paths are included
        env["PATH"] = "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:" + env.get("PATH", "")
        
        process = subprocess.Popen(
            command, shell=True, executable="/bin/bash", stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, text=True, env=env
        )
        
        if wait:
            stdout, stderr = process.communicate()
            stdout = stdout.encode("ascii", "ignore").decode("ascii").strip()
            stderr = stderr.encode("ascii", "ignore").decode("ascii").strip()
            excode = process.returncode
        else:
            stdout, stderr, excode = "", "", None
    except Exception as e:
        stdout, stderr, excode = "", str(e), 9
    
    result = {
        "stdout": stdout,
        "stderr": stderr,
        "excode": excode
    }
    
    return json.dumps(result)

@close_old_connections
def drop_mysql_node(node, final_backup = False):
        cluster_ids = c123_node.objects.filter(cluster = node.cluster, ind_deleted = False).values_list('id', flat=True)
        existing_drop = Task.objects.filter(target_id__in = cluster_ids, task = 'drop_mysql_node_task', status__in = [TaskState.PENDING.value, TaskState.RUNNING.value])
        if existing_drop:
            raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        node.status = Status.DROPPING.value
        node.save()
        if node.cluster_role != ClusterRoles.PRIMARY.value:
            final_backup = False
        if final_backup:
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.MYSQL.value)
            keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
            # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            config = SystemConfig.objects.filter(db_type = DBType.MYSQL.value).first()
            backup_path = config.backups_path
            ex_params = {
                'size': 0,
                'duration': 0,
                'belong_to': DBType.MYSQL.value,
                'node_version': node.version,
                'compress': True,
                'root_pwd': node.ex_params.get('mysql_root_password'),
                'modified_date': '',
                'backup_path': f'{backup_path}/{node.host.name}_{node.port}/{now}'
            }

            final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
            node = node,
            type = Inventory.BACKUP.value,
            created_by = node.created_by,
            is_sched = False,
            keep_days = keep_days,
            ex_params = ex_params
            )

        cleanup_logs(node.pk, node.type)
        update_host_usage(node.host)

        from django.conf import settings
        # check if node the only one in cluster
        other_nodes = c123_node.objects.filter(cluster = node.cluster, type = DBType.MYSQL.value, ind_deleted = False).exclude(pk = node .pk)
        has_other_nodes = other_nodes.exists()
        if not has_other_nodes:
            job = settings.scheduler.get_job(job_id=f"update_mysql_fork_roles_{node.cluster.id}")
            if job:
                job.remove()

        node_jobs = [i for i in settings.scheduler.get_jobs() if node.name in i.id]
        for job in node_jobs:
            job.remove()

        node_data = object_to_json(node)
        host_data = c123_HostSerializer(node.host).data
        created_by = node.created_by
        standbys = [{"node": object_to_json(i), "host": object_to_json(i.host)} for i in other_nodes.filter(cluster_role = ClusterRoles.STANDBY.value)]
        tmp_data = {'node_data': node_data, 'host_data': host_data,'standbys': standbys, 'created_by': created_by.id}
        primary =  c123_node.objects.filter(cluster=node.cluster, type = DBType.MYSQL.value, cluster_role=ClusterRoles.PRIMARY.value, ind_deleted = False).first()
        if primary:
            tmp_data['primary_data'] = object_to_json(primary)
            tmp_data['host_primary'] = c123_HostSerializer(primary.host).data

        tmp_data['final_backup'] = c123_inventory_serializer(final_backup).data if final_backup else final_backup
        tmp_data['backup_host'] = c123_HostSerializer(app_host).data
        tmp_data = json.dumps(tmp_data)
        summary = generate_license_summary()

        from clusterCore.api.licensing.views import LicenseType
        revoke_license(node)
        spread_licenses_by_type(LicenseType.MYSQL.value, summary.mysql)
        task = Task.objects.create(task='drop_mysql_node_task',
                                    displayment='drop {} {}'.format(node.cluster_role.lower(), node.name),
                                    target_type=DBType.MYSQL.value,
                                    target_id=node.id,
                                    org_id=node.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=created_by)
        return task

def is_valid_email(email):
    if not isinstance(email, str) or not email or email.isspace():
        return True
    # Basic email validation using regex
    regex = r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$'
    if not re.match(regex, email):
        return False
    # Verify email address using SMTP
    try:
        domain = email.split('@')[1]
        answers = dns.resolver.query(domain, 'A')
        return True
    except dns.resolver.NXDOMAIN:
        return False
    except dns.resolver.NoAnswer:
        return False
    except dns.resolver.Timeout:
        return False

def send_mail(subject, body, to, reply_to=None, headers=None):
    email = EmailMessage(subject=subject,body=body, to=to, reply_to=reply_to, headers=headers)
    status = email.send()
    return status

def remote_command(client, cmd, wait_for_output) -> dict:
    json_str = {}
    try:
        _, stdout, stderr = client.exec_command(cmd)
        print('remote_command {}'.format(cmd))
        if wait_for_output:
            try:
                strout = stdout.read().decode('ascii').strip()
                strerr = stderr.read().decode('ascii').strip()
            except UnicodeDecodeError:
                strout = stdout.read().decode('utf-8').strip()
                strerr = stderr.read().decode('utf-8').strip()

        json_str['stdout'] = strout
        json_str['stderr'] = strerr
        json_str['excode'] = stdout.channel.recv_exit_status()
        return json.dumps(json_str)
    except Exception as e:
        json_str['stderr'] = e.strerror if hasattr(e, 'strerror') else str(e)
        json_str['stdout'] = ''
        json_str['excode'] = 1
        return json.dumps(json_str)

def mysql_ismaster(node) -> str:


    try:
        error_label = 'mysql_ismaster'
        node_client = create_mysql_connection(node)
        if isinstance(node_client, tuple):
            error = node_client[1].strerror if hasattr(node_client[1], 'strerror') else str(node_client[1])
            return error
        cur = node_client.cursor()
        cur.execute(f"SHOW {'REPLICA' if DBType.MYSQL.value in node.type and float(node.version) > 5.7 else 'SLAVE'} STATUS;")
        res = cur.fetchone()
        node.severity = Severity.OK.value
        node.status = Status.OK.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        return 'no' if res else 'yes'
    except Exception as e:
       error = e.strerror if hasattr(e, 'strerror') else str(e)
       node.severity = Severity.NOT_OK.value
       node.status = Status.DOWN.value
       node.ind_started = True
       node.save()
       c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
       print(error_label, node.name, error)
       return error

def get_mysql_fork_standby_lag(standby) -> str:


    try:

        error_label = f'get_{standby.type.lower()}_lag'
        slave_client = create_mysql_connection(standby)
        if isinstance(slave_client, tuple):
            return "N/A"
        cur = slave_client.cursor(dictionary=True)
        cur.execute(f"SHOW {'REPLICA' if float(standby.version) > 5.7 and DBType.MYSQL.value in standby.type else 'SLAVE'} STATUS;")
        res = cur.fetchone()
        standby.severity = Severity.OK.value
        standby.status = Status.OK.value
        standby.ind_started = True
        standby.save()
        if not res:
            return "N/A"
        lag = res.get(f"Seconds_Behind_{'Source' if float(standby.version) > 5.7 and DBType.MYSQL.value in standby.type else 'Master' }")
        if not lag:
            return "N/A"

        return f'{lag} Seconds'
    except Exception as e:
       standby.severity = Severity.NOT_OK.value
       standby.status = Status.DOWN.value
       standby.ind_started = True
       standby.save()
       error = e.strerror if hasattr(e, 'strerror') else str(e)
       print(error_label, standby.name, error)
       return "N/A"

def find_mongo_bin(host, mongo_version) -> str:
    """find location of mongo bin"""

    mongo_bin = '/usr/bin'
    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            print('find_mongo_bin:', client[1].strerror)
        print('find_mongo_bin:',client[1])
        return mongo_bin

    jsn = json.loads(remote_command(client, f'sudo dirname $(rpm -ql mongodb-org-server-{mongo_version}.*-1.el7.x86_64|grep /bin/mongod|grep -v share)', True), strict=False)
    # jsn = json.loads(remote_command(host, 'which mongo', True), strict=False)
    if jsn['stdout'] != '':
        mongo_bin = jsn['stdout'].strip()

    if jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower() and 'permission denied' not in jsn['stderr'].lower():
        if 'not found' in jsn['stderr'].lower() or 'dirname --help' in jsn['stderr'].lower():
            mongo_bin = '/usr/bin'

        else:
            print('find_mongo_bin:', jsn['stderr'])

    return mongo_bin

def run_python_script(script_path, script_args):
    """
    Executes a Python script with the given arguments.

    Parameters:
        script_path (str): Path to the Python script.
        script_args (list): List of arguments to pass to the script.

    Returns:
         dict[str, Any]: dictionary with keys 'stdout', 'stderr', and 'excode'.
    """
    try:
        # Call the script and capture stdout, stderr
        result = subprocess.run(
            [sys.executable, script_path] + script_args,
            text=True,
            capture_output=True
        )
        # Return stdout, stderr, and exit code in JSON format
        return {
            "stdout": result.stdout,
            "stderr": result.stderr,
            "excode": result.returncode
        }
    except Exception as e:
        # Handle unexpected exceptions
        return {
            "stdout": "",
            "stderr": str(e),
            "excode": -1
        }

def check_space(node, path, source_path = None, required_space = None):
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")

    host = node.host if node.type not in [DBType.MSSQL.value, DBType.POSTGRES_K8S.value] else app_host

    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])

    jsn = json.loads(remote_command(client, f"sudo stat '{path}'", True), strict = False)
    if jsn['stderr'] != '':
        if "cannot stat" in jsn["stderr"].lower():
            jsn = json.loads(remote_command(client, f"sudo mkdir -p '{path}'", True), strict=False)
            error = str(jsn['stderr'])
            if error != '':
                raise APIException400(error)

    cmd = f'''df -k "{path}" | awk 'NR>1 {{print $4}}' '''
    jsn = json.loads(remote_command(client, cmd, True), strict=False)

    if jsn['stderr'] != '':
            raise APIException400(jsn['stderr'])

    available_space = jsn['stdout']


    # case when backup storage have different os family than node machine
    if not required_space:
        match node.type:
            case DBType.MSSQL.value:
                conn = connect_to_sqlserver_instance(node)
                if isinstance(conn, tuple):
                    if hasattr(conn[1], 'strerror'):
                        raise APIException401(conn[1].strerror)
                    raise APIException401(str(conn[1]))
                else:
                    try:
                        with conn:
                            with conn.cursor() as cur:
                                cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
                                res = cur.fetchone()
                                if not res or not res[0]:
                                    raise APIException500("Could not obtain required space for backup.")
                                required_space = float(res[0])
                    except Exception as error:
                        raise APIException500(str(error))
            case _ :
                if source_path:
                    client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key, node.host.ssh_port)
                    if isinstance(client, tuple):
                        if hasattr(client[1], 'strerror'):
                            raise APIException401(client[1].strerror)
                        raise APIException401(client[1])
                    jsn = json.loads(remote_command(client, f"sudo du -s -BK '{source_path}' | awk '{{print $1}}' | tr -d 'K'", True), strict = False)
                    if jsn['stderr'] != '':
                        raise APIException400(jsn['stderr'])
                    required_space = float(jsn['stdout'])

    if not source_path:
        if float(available_space) > float(required_space):
            return {'available_space': available_space, 'status': 200}
        else:
            return {'available_space': available_space, 'status': 406}
    else:
        if float(available_space) > float(required_space):
            return {'available_space': available_space, 'required_space': required_space, 'status': 200}
        else:
            raise APIException406(f'{path} has {available_space} KB free space. {required_space} KB required')

@close_old_connections
def drop_backup(backup_list, automate = False):
    task = SimpleNamespace(**{"displayment": "no supported backups found"})
    if not backup_list.exists():
        return task

    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not find the app host")

    backup_host = c123_HostSerializer(app_host).data
    data_backups = backup_list.exclude(type = Inventory.FLASHBACK.value)
    restore_points =  backup_list.filter(type = Inventory.FLASHBACK.value)
    for restore_point in restore_points:
        node = restore_point.node
        match node.type:
            case DBType.ORACLE_PDB.value:
                cdb = c123_node.objects.filter(type = DBType.ORACLE_CDB.value, ind_deleted = False, ex_params__cdb_id = node.ex_params.get('cdb_id')).first()
                if cdb:
                    tmp_data = json.dumps({'cdb_data': object_to_json(cdb), 'host_data': host_data, 'backup_host': backup_host, 'flashback_data': object_to_json(restore_point), 'pdb_name': node.name, 'pdb_id': node.id, 'automate': automate, 'created_by': restore_point.created_by})
                    #problem - it expects to get pdb flashback, but we also have cdb flashbacks
                    task = Task.objects.create(task='drop_restore_point_task',
                                    displayment=f'drop flashback {restore_point.name}',
                                    target_type=DBType.ORACLE_PDB.value,
                                    target_id=node.id,
                                    org_id=node.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=restore_point.created_by)
            case _:
                pass

    if data_backups.exists():
        backup_data = []
        for backup in  data_backups:
            if backup.node and backup.is_sched:
                type = backup.ex_params.get('belong_to')
                from django.conf import settings
                job = settings.scheduler.get_job(job_id= str(backup.ex_params.get('job_id')))
                if job:
                    job.remove()

                match type:
                    case DBType.MONGO.value:
                        job = settings.scheduler.get_job(job_id= str(backup.ex_params.get('oplog_job_id')))
                        if job:
                            job.remove()
                        oplog_dump_tasks = Task.objects.filter(task ='mongodump_oplog_task', pk = backup.ex_params.get('oplog_task_id'))
                        if oplog_dump_tasks.exists():
                            oplog_dump_tasks.delete()
                    case DBType.POSTGRES.value:
                        if backup.is_sched and backup.ex_params.get('incremental'):
                            node = backup.node
                            if node:
                                cluster = node.cluster
                                if cluster:
                                    if not cluster.ex_params:
                                        cluster.ex_params = {}
                                    #if we removing incremental backup -
                                    # then need reset new_basebackup
                                    cluster.ex_params['new_basebackup'] = True
                                    cluster.save()
                    case _:
                        pass

            if backup.ex_params:
                backup_tasks = Task.objects.filter(pk = backup.ex_params.get('task_id'))
                if backup_tasks.exists():
                    backup_tasks.delete()

            node = backup.node if backup.node else backup.database.node
            host_data = c123_HostSerializer(node.host).data
            backup_data.append({"backup": object_to_json(backup), "host": host_data})

        first_backup = data_backups.first()
        tmp_data = json.dumps({'backup_data': json.dumps(backup_data), 'backup_host': backup_host, 'automate': automate,  'created_by': first_backup.created_by.id})
        task = Task.objects.create(task='drop_backup_task',
                                    displayment='drop backup',
                                    target_type=first_backup.node.type,
                                    target_id=first_backup.id,
                                    org_id=first_backup.node.org_id,
                                    schedule=Schedule.NOW.value,
                                    status=TaskState.PENDING.value,
                                    parameters=tmp_data,
                                    created_by=first_backup.created_by)
    if not automate:
        backup_list.update(is_sched = False)

    return task

def connect_to_redis(node):
    """connecting to redis cluster"""

    try:

        result = check_host_connection(node)
        host = node.host
        client = Redis(host=host.name, port=node.port, username=node.database_user, password=node.database_pwd, socket_connect_timeout=Timeout.BACKEND_REDIS.value, retry_on_timeout=True)
        client.ping()
        return client
    except Exception as e:
        if not isinstance(result, tuple):
            if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                node.status = Status.STOPPED.value
            else:
                node.status = Status.DOWN.value

            node.severity = Severity.NOT_OK.value
            node.ind_started = False
            node.save()
            c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        logging.error(f"create_redis_connection: {e}")
        return (False, e)

def check_redis_role(node):


    try:
        redis_client = connect_to_redis(node.host.name, node.database_user, node.database_pwd, node.port)
        if isinstance(redis_client, tuple):
            error = redis_client[1].strerror if hasattr(redis_client[1], 'strerror') else redis_client[1]
            return error
        role = redis_client.info().get("role")
        return role
    except Exception as e:
        node.severity = Severity.NOT_OK.value
        node.status = Status.DOWN.value
        node.ind_started = True
        node.save()
        error = e.strerror if hasattr(e, 'strerror') else str(e)
        print("check_redis_role:", error)
        return error

def get_node_os(host_data) -> str:
    try:
        client = create_ssh_connection(host_data['name'], host_data['username'], host_data['password'], host_data['private_key'], host_data['ssh_port'])
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                raise APIException403(client[1].strerror)
            raise APIException403(client[1])
        # centos_dist = False
        jsn = json.loads(remote_command(client, f"if [ -f '/etc/centos-release' ]; then echo '{OperatingSystem.CENTOS.value}'; elif [ -f '/etc/lsb-release' ]; then echo '{OperatingSystem.UBUNTU.value}'; else echo '{OperatingSystem.UNKNOWN.value}'; fi", True), strict=False)
        if jsn['stderr'] != '':
            raise APIException500(jsn['stderr'])

        if OperatingSystem.UNKNOWN._missing_ in jsn['stdout']:
            return (OperatingSystem.UNKNOWN.value, 'Did not find a distributor')

        centos_dist = True if OperatingSystem.CENTOS.value in jsn['stdout'] else False
        distributor = OperatingSystem.CENTOS.value if centos_dist else OperatingSystem.UBUNTU.value
        return distributor
    except Exception as error:
        return (OperatingSystem.INVALID.value, str(error))

def check_os_version(host) -> tuple:
    """return distributor and release version"""
    try:
        client = create_ssh_connection(host['name'], host['username'], host['password'], host['private_key'], host['ssh_port'])
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                error = client[1].strerror
            else:
                error = str(client[1])
            return (OperatingSystem.INVALID.value, {'stderr': error})

        jsn = json.loads(remote_command(client, "source /etc/os-release; echo $ID", True), strict=False)
        if jsn['stderr'] != '':
            return ('error', jsn)

        distributor = jsn['stdout'].strip()

        if any(v in distributor for v in [OperatingSystem.UBUNTU.value, OperatingSystem.DEBIAN.value]):
            distributor = OperatingSystem.UBUNTU.value
        elif any(v in distributor for v in [OperatingSystem.CENTOS.value, OperatingSystem.REDHAT.value, OperatingSystem.FEDORA.value, OperatingSystem.ROCKY.value]):
            distributor = OperatingSystem.CENTOS.value
        else:
            return (OperatingSystem.UNKNOWN.value, {'stdout': '', 'stderr': f'Unsupported OS: {distributor}', 'excode': 1})

        jsn = json.loads(remote_command(client,"source /etc/os-release; echo $VERSION_ID", True), strict=False)
        if jsn['stderr'] != '':
            return (OperatingSystem.INVALID.value, jsn)

        release = jsn['stdout'].strip().split('.')[0]


        return (distributor, release)
    except Exception as error:
        jsn = {'stdout': '', 'stderr': str(error), 'excode': 9}
        return (OperatingSystem.INVALID.value, jsn)

def get_pg_standby_lag(standby) -> str:

        # check if instance is down


        try:
            conn = get_postgres_connection(standby)
            if isinstance(conn, tuple):
                return "N/A"

            # check if repmgr is absent
            conn = get_postgres_connection(standby, 'repmgr')
            if isinstance(conn, tuple):
                standby.severity = Severity.WARNING.value
                standby.ind_started = True
                standby.status = Status.NO_REPMGR.value
                standby.save()
                error = conn[1].strerror if hasattr(conn[1], 'strerror') else str(conn[1])
                logging.error(f'get_standby_lag: {error}')
                return "N/A"
            cur = conn.cursor()
            cur.execute("SELECT replication_lag FROM repmgr.replication_status")
            res = cur.fetchone()
            standby.severity = Severity.OK.value
            standby.status = Status.OK.value
            standby.ind_started = True
            standby.save()
            if res:
                return res[0]
        except Exception as e:
            standby.severity = Severity.NOT_OK.value
            standby.status = Status.DOWN.value
            standby.save()
            error = e.strerror if hasattr(e, 'strerror') else str(e)
            logging.error(f'get_standby_lag: {error}')
            return "N/A"

        client = create_ssh_connection(standby.host.name, standby.host.username, standby.host.password, standby.host.private_key, standby.host.ssh_port)
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                logging.error(f'get_standby_lag: {client[1].strerror}')
            else:
                logging.error(f'get_standby_lag: {client[1]}')
            return "N/A"

        postgres_bin = find_postgres_bin(client, standby.version)
        repmgr_pwd = standby.ex_params.get("repmgr_pwd")

        jsn = json.loads(remote_command(client, f"sudo -iu postgres PGPASSWORD='{repmgr_pwd}' {postgres_bin}/repmgr -f '/etc/repmgr/{standby.version}/repmgr.conf' node check", True), strict=False)
        if jsn['stderr'] != '' and "error" in jsn["stderr"].lower():
            standby.severity = Severity.NOT_OK.value
            standby.status = Status.DOWN.value
            standby.ind_started = False
            standby.save()
            logging.error(f"get_standby_lag: {jsn['stderr']}")
            return "N/A"

        standby.severity = Severity.OK.value
        standby.status = Status.OK.value
        standby.ind_started = True
        standby.save()

        try:
            lag = [i for i in jsn['stdout'].split('\n') if 'Replication lag:' in i][0].split(':')[1].strip()
        except Exception as error:
            logging.error(f"get_standby_lag: {str(error)}")
            return "N/A"

        return lag

def get_mongo_standby_lag(standby) -> str:


    try:
        client = create_mongo_connection(standby)
        if isinstance(client, tuple):
            return "N/A"
        db_stats = client.admin.command({'replSetGetStatus'  :1})
    except Exception as e:
        standby.severity = Severity.NOT_OK.value
        standby.status = Status.DOWN.value
        standby.ind_started = True
        standby.save()
        error = e.details.get("errmsg")
        logging.error(f'get_standby_lag: {error}')
        return "N/A"

    standby.severity = Severity.OK.value
    standby.status = Status.OK.value
    standby.ind_started = True
    standby.save()
    primary = next((i for i in db_stats["members"] if i["stateStr"] == ClusterRoles.PRIMARY.value), None)
    if primary:
        primary_optime = primary["optimeDate"]
        secondary = next((i for i in db_stats["members"] if i["stateStr"] == "SECONDARY" and i["name"] == standby.name), None)
        if secondary:
            secondary_optime = secondary["optimeDate"]
            lag = abs((primary_optime - secondary_optime).total_seconds())
            return f"{lag} Seconds"
        else:
            return "N/A"
    else:
        return "N/A"

def check_port_existence(host, port, type = None, node = None) -> bool:
    """check if the port is not in use,
    return False if not in use otherwise True"""
    match type:
        case DBType.MSSQL.value:
            client = connect_to_sqlserver_instance(node)
            if isinstance(client, tuple):
                if hasattr(client[1], 'strerror'):
                    raise APIException401(client[1].strerror)
                raise APIException401(client[1])

            try:
                cursor = client.cursor()
                cursor.execute(f"SELECT port FROM sys.tcp_endpoints WHERE port=?", [port])
                result = cursor.fetchall()
            except Exception as e:
                error =  e.strerror if hasattr(e, 'strerror') else str(e)
                raise APIException400(error)

            return len(result) > 0

        case _:
            client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
            if isinstance(client, tuple):
                if hasattr(client[1], 'strerror'):
                    raise APIException401(client[1].strerror)

                raise APIException401(client[1])

            jsn = json.loads(remote_command(client,  f"sudo netstat -tunlp|grep -w '{port}'", True), strict=False)
            if jsn['stdout'] != '':
                return True

            if jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower():
                raise APIException500(jsn['stderr'])

            return False

def find_tilda(host) -> str:
    """find location of ~"""
    client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
    if isinstance(client, tuple):
        if hasattr(client[1], 'strerror'):
            raise APIException401(client[1].strerror)

        raise APIException401(client[1])

    jsn = json.loads(remote_command(client, "sudo cd ~;pwd | tr -d '\\n'", True), strict=False)
    if jsn['stdout'] != '':
        print(jsn['stdout'])

    if jsn['stderr'] != '' and "warning" not in jsn["stderr"].lower():
        raise APIException500(jsn['stderr'])

    return jsn['stdout']

def find_real_pg_primary(app_primary,  silent = True):
    is_primary = check_if_pg_primary(app_primary)
    if is_primary is True:
        return app_primary
    else:
        standbys = c123_node.objects.filter(type = DBType.POSTGRES.value, cluster_role = ClusterRoles.STANDBY.value, cluster  = app_primary.cluster, ind_deleted = False)
        for standby in standbys:
            is_standby_primary = check_if_pg_primary(standby)
            if is_standby_primary is True:
                other_standbys = c123_node.objects.filter(type = DBType.POSTGRES.value, cluster_role = ClusterRoles.STANDBY.value, ind_deleted = False, cluster = app_primary.cluster).exclude(pk = standby.pk)
                for other in other_standbys:
                    other.ex_params['pg_primary_id'] = standby.pk
                    other.save()

                app_primary.cluster_role = ClusterRoles.STANDBY.value
                app_primary.ex_params['pg_primary_id'] = standby.pk
                app_primary.save()
                standby.cluster_role = ClusterRoles.PRIMARY.value
                standby.ex_params['pg_primary_id'] = standby.pk
                standby.save()
                return standby
        else:
            if silent:
                # assume that real primary is last known
                return app_primary
            else:
                raise APIException410("cluster is broken, no found primary")

def update_host_usage(host):
    host.ind_deleted = False if c123_node.objects.filter(host = host, ind_deleted = False).exclude(status = Status.DROPPING.value).exists() else True
    host.save()

def check_if_pg_primary(node, in_cluster=False):
    '''returns True if primary, otherwise - False'''



    try:

        conn = get_postgres_connection(node)
        if isinstance(conn, tuple):
            return False
        cur = conn.cursor()
        cur.execute("SELECT pg_is_in_recovery()")
        node.ind_started = True
        node.severity = Severity.OK.value if node.status == Status.DOWN.value else node.severity
        node.status = Status.OK.value if node.status == Status.DOWN.value else node.status
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        is_primary = not cur.fetchone()[0]
        if in_cluster:
            # Execute query to check if the database exists
            cur.execute("SELECT datname FROM pg_database WHERE datname = %s", ('repmgr',))
            exists = cur.fetchone() is not None
            if not exists:
                node.severity = Severity.WARNING.value
                node.status = Status.NO_REPMGR.value
                node.save()

        return is_primary
    except Exception as e:
        node.severity = Severity.NOT_OK.value
        node.status = Status.DOWN.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        error = e.strerror if hasattr(e, 'strerror') else str(e)
        print("check_if_pg_primary:", error)
        return False

def check_if_pg_standalone(node):
    '''returns False if not standalone, otherwise - True'''



    try:
        conn = get_postgres_connection(node)
        if isinstance(conn, tuple):
            return True
        cur = conn.cursor()
        cur.execute("SELECT datname from pg_database where datname = 'repmgr'")
        node.ind_started = True
        node.severity = Severity.OK.value if node.status == Status.DOWN.value else node.severity
        node.status = Status.OK.value if node.status == Status.DOWN.value else node.status
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        res = cur.fetchone()
        return res != None
    except Exception as e:
        node.severity = Severity.NOT_OK.value
        node.status = Status.DOWN.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        error = e.strerror if hasattr(e, 'strerror') else str(e)
        print("check_if_pg_standalone:", error)
        return True

def get_cron_schedule(hours = None, minutes = None, seconds = None, days_of_week = None, days_of_month = None):
    if not isinstance(hours, int) or not isinstance(minutes, int):
        raise APIException400('unsupported schedule format')

    from django.conf import settings
    if days_of_week and days_of_month:
        return CronTrigger(day_of_week= ",".join(days_of_week), day =",".join(map(str, days_of_month)), hour = hours, minute = minutes, timezone=settings.TIME_ZONE)

    if days_of_week:
        return CronTrigger(day_of_week= ",".join(days_of_week), hour = hours, minute = minutes, timezone=settings.TIME_ZONE)

    if days_of_month:
        return CronTrigger(day =",".join(map(str, days_of_month)), hour = hours, minute = minutes, timezone=settings.TIME_ZONE)

    return CronTrigger(hour = hours, minute = minutes, second = seconds, timezone=settings.TIME_ZONE)

def checkPathExists(host, path, host_type = None) -> bool:
    """check if a directory exists on remote host"""
    if host_type == 'windows':
        client = create_winrm_connection(host.name, host.username, host.password, host.ssh_port)
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                raise APIException401(client[1].strerror)

            raise APIException401(client[1])

        jsn = json.loads(remote_command_win(client,f'dir "{path}"', True), strict=False)
        if jsn['stderr'] :
            if any(err in jsn['stderr'].lower() for err in ["cannot find", "not found"]):
                return False

            raise APIException500(jsn['stderr'])

        return True
    else:
        client = create_ssh_connection(host.name, host.username, host.password, host.private_key, host.ssh_port)
        if isinstance(client, tuple):
            if hasattr(client[1], 'strerror'):
                raise APIException401(client[1].strerror)
            raise APIException401(client[1])

        jsn = json.loads(remote_command(client, f"sudo stat '{path}'", True), strict = False )
        if jsn['stderr'] != '':
            if "cannot stat" in jsn["stderr"].lower():
                return False
            else:
                raise APIException500(jsn['stderr'])

        return True

def check_if_elastic_primary(node):
    es_client = create_elasticsearch_insecure_connection(node)

    if isinstance(es_client, tuple):
        return False

    try:
        res = es_client.cat.master(format="json")
        if len(res) > 0:
            return res[0].get('node') == node.name
        return False
    except Exception as e:
        node.severity = Severity.NOT_OK.value
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        node.status = Status.DOWN.value
        node.ind_started = True
        node.save()
        print("check_if_elastic_primary:", node.name, str(e))
        return False
